/*									HTTCP.c
**	GENERIC COMMUNICATION CODE
**
**	(c) COPYRIGHT CERN 1994.
**	Please first read the full copyright statement in the file COPYRIGH.
**
**	This code is in common between client and server sides.
**
**	16 Jan 92  TBL	Fix strtol() undefined on CMU Mach.
**	25 Jun 92  JFG  Added DECNET option through TCP socket emulation.
**	13 Sep 93  MD   Added correct return of vmserrorno for HTInetStatus.
**			Added decoding of vms error message for MULTINET.
**	31 May 94  HF	Added cache on host id's; now use inet_ntoa() to
**			HTInetString and some other fixes. Added HTDoConnect
**			and HTDoAccept
*/

/* Library include files */
#include "RoverBase.h"
#include "tcp.h"
#include "HTUtils.h"
#include "HTString.h"
#include "HTAtom.h"
#include "HTList.h"
#include "HTParse.h"
#include "HTAccess.h"
#include "HTError.h"
#include "HTThread.h"
#include "HTTCP.h"					 /* Implemented here */
#include "RoverHost.h"
#include "RoverTCP.h"

#ifdef VMS 
#include "HTVMSUtils.h"
#endif /* VMS */

#ifdef sparc
#include "netinet/tcp.h"
#endif /* sparc */

#ifdef SHORT_NAMES
#define HTErrnoString		HTErrnoS
#define HTInetString 		HTInStri
#define HTParseInet		HTPaInet
#endif


/* VMS stuff */
#ifdef VMS
#ifndef MULTINET
#define FD_SETSIZE 32
#else /* Multinet */
#define FD_SETSIZE 256
#endif /* Multinet */
#endif /* VMS */

/* x seconds penalty on a multi-homed host if IP-address is down */
#define TCP_PENALTY		1200

/* x seconds penalty on a multi-homed host if IP-address is timed out */
#define TCP_DELAY		600

/* Globals */
PUBLIC unsigned int	ConCacheSize = 512;	 /* Number of cached servers */

/* Type definitions and global variables etc. local to this module */

/* This structure is a cache of hosts to whom we have connected over time.
   The structure contains the necessary parts from hostent. For Internet host
   hostent->h_addr_list is not an array of char pointers but an array of 
   pointers of type in_addr. */
typedef struct _host_info {
    HTAtom *		hostname;		    /* Official name of host */
    int			hits;		/* Total number of hits on this host */
    int			addrlength;	       /* Length of address in bytes */
    int			homes;	       /* Number of IP addresses on the host */
    int			offset;         /* Offset value of active IP address */
    char **		addrlist;      /* List of addresses from name server */
    float *		weight;			   /* Weight on each address */
} host_info;

PRIVATE HTList *hostcache = NULL;  /* List of servers that we have talked to */
PRIVATE unsigned int CacheSize = 0;		    /* Current size of cache */

/* ------------------------------------------------------------------------- */

/* Close a socket and update the list of open sockets (used by clients when
 *   they fork.
 */
PUBLIC int RoverSockClose(int sockfd)
{
  if (RoverDebug) fprintf(stderr, ">>> CLOSE %d\n", sockfd);
  return(close(sockfd));
}

/*
**	Returns the string equivalent to the errno passed in the argument.
**	We can't use errno directly as we haev both errno and socerrno. The
**	result is a static buffer.
*/
PUBLIC CONST char *HTErrnoString ARGS1(int, errornumber)
{
#if defined(NeXT) || defined(THINK_C)
    return strerror(errornumber);
#else
#ifdef VMS
    static char buf[60];
    sprintf(buf,"Unix errno = %ld dec, VMS error = %lx hex", errornumber,
	    vaxc$errno);
    return buf;
#else
    return (errornumber < sys_nerr ? sys_errlist[errornumber]:"Unknown error");
#endif
#endif
}


/* ------------------------------------------------------------------------- */
/*	       		     HOST CACHE MANAGEMENT 			     */
/* ------------------------------------------------------------------------- */

/*                                                     	TCPCacheRemoveElement
**
**	Remove the element specified from the cache
*/
PRIVATE void TCPCacheRemoveElement ARGS1(host_info *, element)
{
    if (!hostcache) {
        if (PROT_TRACE)
            fprintf(TDEST, "HostCache... Remove not done, no cache\n");
        return;
    }
    if (PROT_TRACE) fprintf(TDEST, "HostCache... Remove `%s' from cache\n",
			    HTAtom_name(element->hostname));
    HTList_removeObject(hostcache, (void *) element);
    if (*element->addrlist)
	free(*element->addrlist);
    if (element->addrlist)
	free(element->addrlist);
    if (element->weight)
	free(element->weight);
    free(element);
}


/*                                                     	TCPCacheRemoveAll
**
**	Cleans up the memory. Called by HTLibTerminate
*/
PUBLIC void TCPCacheRemoveAll NOARGS
{
    if (hostcache) {
	HTList *cur = hostcache;
	host_info *pres;
	while ((pres = (host_info *) HTList_nextObject(cur))) {
	    if (*pres->addrlist)
		free(*pres->addrlist);
	    if (pres->addrlist)
		free(pres->addrlist);
	    if (pres->weight)
		free(pres->weight);
	    free(pres);
	}
	HTList_delete(hostcache);
	hostcache = NULL;
    }
}


/*                                                     TCPCacheRemoveHost
**
**	Removes the corresponding entrance in the cache
*/
PRIVATE void TCPCacheRemoveHost ARGS1(char *, host)
{
    HTAtom *hostatom = HTAtom_for(host);
    HTList *cur = hostcache;
    host_info *pres = NULL;
    if (!hostcache) {
	if (PROT_TRACE)
	    fprintf(TDEST, "HostCache... Remove host not done, no cache\n");
	return;
    }
    while ((pres = (host_info *) HTList_nextObject(cur)) != NULL) {
	if (pres->hostname == hostatom) {
	    break;
	}
    }
    if (pres)
	TCPCacheRemoveElement(pres);
}

/* Maximum value an `unsigned int' can hold.  (Minimum is 0.)  */
#ifndef UINT_MAX
#ifdef  __STDC__
#define UINT_MAX        4294967295U
#else
#define UINT_MAX        4294967295
#endif
#endif /* !UINT_MAX */

/*                                                     TCPCacheGarbage
**
**	Remove the element with the lowest hit rate
*/
PRIVATE void TCPCacheGarbage NOARGS
{
    HTList *cur = hostcache;
    host_info *pres, *worst_match = NULL;
    unsigned int worst_hits = UINT_MAX;
    if (!hostcache) {
	if (PROT_TRACE)
	    fprintf(TDEST, "HostCache... Garbage collection not done, no cache\n");
	return;
    }

    /* Seek for worst element */
    while ((pres = (host_info *) HTList_nextObject(cur))) {
	if (!worst_match || pres->hits <= worst_hits) {
	    worst_match = pres;
	    worst_hits = pres->hits;
	}
    }
    if (worst_match)
	TCPCacheRemoveElement(worst_match);
}


/*                                                     TCPCacheAddElement
**
**	Add an element to the cache of visited hosts. Note that this function
**	requires the system implemented structure hostent and not our own
**	host_info. The homes variable indicates the number of
**	IP addresses found.
**
**      Returns new element if OK NULL if error
*/
PRIVATE host_info *TCPCacheAddElement ARGS2(HTAtom *, host,
					    struct hostent *, element)
{
    host_info *newhost;
    char *addr;
    char **index = element->h_addr_list;
    int cnt = 1;
    if (!host || !element) {
	if (PROT_TRACE)
	    fprintf(TDEST, "HostCache... Bad argument to add to cache\n");
	return NULL;
    }
    while(*index++)
	cnt++;
    if ((newhost = (host_info *) calloc(1, sizeof(host_info))) == NULL ||
	(newhost->addrlist = (char **) calloc(1, cnt*sizeof(char*))) == NULL ||
	(addr = (char *) calloc(1, cnt*element->h_length)) == NULL)
	outofmem(__FILE__, "TCPCacheAddElement");
    newhost->hostname = host;
    index = element->h_addr_list;
    cnt = 0;
    while (*index) {
	*(newhost->addrlist+cnt) = addr+cnt*element->h_length;
	memcpy((void *) *(newhost->addrlist+cnt++), *index++,
	       element->h_length);
    }
    newhost->homes = cnt;
    if ((newhost->weight = (float *) calloc(newhost->homes,
					    sizeof(float))) == NULL)
	outofmem(__FILE__, "TCPCacheAddElement");

    newhost->addrlength = element->h_length;
    if (!hostcache)
	hostcache = HTList_new();

    if (PROT_TRACE) {
	if (newhost->homes == 1)
	    fprintf(TDEST, "HostCache... Adding single-homed host `%s'\n",
		    HTAtom_name(host));
	else
	    fprintf(TDEST, "HostCache... Adding host `%s' with %d homes\n",
		    HTAtom_name(host), newhost->homes);
    }
    HTList_addObject(hostcache, (void *) newhost);
    CacheSize++;				/* Update number of elements */
    return newhost;
}


/*                                                     	      TCPAddrWeights
**
**	This function calculates the weights of the different IP addresses
**	on a multi homed host. Each weight is calculated as
**
**		w(n+1) = w(n)*a + (1-a) * deltatime
**		a = exp(-1/Neff)
**		Neff is the effective number of samples used
**		deltatime is time spend on making a connection
**
**	A short window (low Neff) gives a high sensibility, but this is
**	required as we can't expect a lot of data to test on.
**
*/
PUBLIC void TCPAddrWeights ARGS2(char *, host, time_t, deltatime)
{
    HTAtom *hostatom = HTAtom_for(host);
    HTList *cur = hostcache;
    host_info *pres = NULL;
    if (!hostcache) {
	fprintf(TDEST, "HostCache... Weights not calculated, no cache\n");
	return;
    }
    /* Skip any port number from host name */
    if (strchr(host, ':')) {
	char *newhost = NULL;
	char *strptr;
	StrAllocCopy(newhost, host);
	strptr = strchr(newhost, ':');
	*strptr = '\0';
	hostatom = HTAtom_for(newhost);
	free(newhost);
    } else
	hostatom = HTAtom_for(host);
    
    while ((pres = (host_info *) HTList_nextObject(cur)) != NULL) {
	if (pres->hostname == hostatom) {
	    break;
	}
    }
    if (pres) {
	int cnt;
	CONST float passive = 0.9; 	  /* Factor for all passive IP_addrs */
#if 0
	CONST int Neff = 3;
	CONST float alpha = exp(-1.0/Neff);
#else
	CONST float alpha = 0.716531310574;	/* Doesn't need the math lib */
#endif
	for (cnt=0; cnt<pres->homes; cnt++) {
	    if (cnt == pres->offset) {
		*(pres->weight+pres->offset) = *(pres->weight+pres->offset)*alpha + (1.0-alpha)*deltatime;
	    } else {
		*(pres->weight+cnt) = *(pres->weight+cnt) * passive;
	    }
	    if (PROT_TRACE)
		fprintf(TDEST, "AddrWeights. Home %d has weight %4.2f\n", cnt,
			*(pres->weight+cnt));
	}
    } else if (PROT_TRACE) {
	fprintf(TDEST, "HostCache... Weights not calculated, host not found in cache: `%s\'\n", host);
    }
}

/* ------------------------------------------------------------------------- */
/*	       		     HOST NAME FUNCTIONS 			     */
/* ------------------------------------------------------------------------- */

#ifndef DECNET  /* Function only used below for a trace message */

/*	Produce a string for an Internet address
**	----------------------------------------
**
** On exit,
**	returns	a pointer to a static string which must be copied if
**		it is to be kept.
*/

PUBLIC CONST char * HTInetString ARGS1(SockA *, sin)
{
#if 0
    /* This dumps core on some Sun systems :-(. The problem is now, that 
       the current implememtation only works for IP-addresses and not in
       other address spaces. */
    return inet_ntoa(sin->sin_addr);
#endif
    static char string[16];
    sprintf(string, "%d.%d.%d.%d",
	    (int)*((unsigned char *)(&sin->sin_addr)+0),
	    (int)*((unsigned char *)(&sin->sin_addr)+1),
	    (int)*((unsigned char *)(&sin->sin_addr)+2),
	    (int)*((unsigned char *)(&sin->sin_addr)+3));
    return string;
}
#endif /* Decnet */


/*	                                                     HTGetHostByName
**
**	Searched first the local cache then asks the DNS for an address of
**	the host specified.
**
**      Returns:	>0 if OK the number of homes are returned
**		     	-1 if error
*/
PUBLIC int HTGetHostByName ARGS3(char *, host, SockA *, sin,
				 BOOL, use_cur)
{
    HTAtom *hostatom = HTAtom_for(host);
    host_info *pres = NULL;
    if (!hostcache)
	hostcache = HTList_new(); 	     	       /* First time through */
    else {
	HTList *cur = hostcache;       		 	     /* Search cache */
	while ((pres = (host_info *) HTList_nextObject(cur)) != NULL) {
	    if (pres->hostname == hostatom) {
		if (PROT_TRACE)
		    fprintf(TDEST, "HostByName.. Host `%s\' found in cache.\n", host);
		break;
	    }
	}
    }
    
    /* If the host was not found in the cache, then do gethostbyname.
       If we are talking to a multi homed host then take the IP address with
       the lowest weight. If `use_cur'=YES then use current IP-address */
    if (pres) {
	if (pres->homes > 1 && !use_cur) {
	    int cnt;
	    float best_weight = 1e30;		    /* Should be FLT_MAX :-( */
	    for (cnt=0; cnt<pres->homes; cnt++) {
		if (*(pres->weight+cnt) < best_weight) {
		    best_weight = *(pres->weight+cnt);
		    pres->offset = cnt;
		}
	    }
	}
	pres->hits++;	 	 /* Update total number of hits on this host */
    } else {						/* Go and ask for it */
	struct hostent *hostelement;			      /* see netdb.h */
#ifdef MVS	/* Outstanding problem with crash in MVS gethostbyname */
	if (PROT_TRACE)
	    fprintf(TDEST, "TCP on MVS gethostbyname(%s)\n", host);
#endif
	if ((hostelement = gethostbyname(host)) == NULL) {
	    if (PROT_TRACE)
		fprintf(TDEST, "HostByName.. Can't find internet node name `%s'.\n", host);
	    return -1;
	}
	
	/* Add element to the cache and maybe do garbage collection */
	if (CacheSize >= ConCacheSize)
	    TCPCacheGarbage();
	if ((pres = TCPCacheAddElement(hostatom, hostelement)) == NULL) {
	    return -1;
	}
    }
    
    /* Update socket structure using the element with the lowest weight. On
       single homed hosts it means the first value */
    memcpy(&sin->sin_addr, *(pres->addrlist+pres->offset), pres->addrlength);
    return pres->homes;
}


/*	Parse a network node address and port
**	-------------------------------------
**
** On entry,
**	str	points to a string with a node name or number,
**		with optional trailing colon and port number.
**	sin	points to the binary internet or decnet address field.
**
** On exit,	-1	If error
**		>0	If OK the number of homes on the host
**	*sin	is filled in. If no port is specified in str, that
**		field is left unchanged in *sin.
**
** NOTE:	It is assumed that any portnumber and numeric host address
**		is given in decimal notation. Separation character is '.'
*/
PUBLIC int HTParseInet ARGS3(SockA *, sin, CONST char *, str,
			     BOOL, use_cur)
{
    char *host = NULL;
    int status = 0;
    StrAllocCopy(host, str);		      /* Take a copy we can mutilate */

    /* Parse port number if present. */    
    {
	char *port;
	if ((port=strchr(host, ':'))) {
	    *port++ = 0;			       	    /* Chop off port */
	    if (isdigit(*port)) {

#ifdef DECNET
		sin->sdn_objnum = (unsigned char)(strtol(port, (char**)0, 10));
#else /* Internet */
		sin->sin_port = htons(atol(port));
#endif
	    } else {
		if (PROT_TRACE)
		    fprintf(TDEST, "ParseInet... No port indicated\n");
		free(host);
		return -1;
	    }
	}
    }

    /* Parse Internet host */
#ifdef DECNET
    /* read Decnet node name. @@ Should know about DECnet addresses, but it's
       probably worth waiting until the Phase transition from IV to V. */

    sin->sdn_nam.n_len = min(DN_MAXNAML, strlen(host));  /* <=6 in phase 4 */
    strncpy (sin->sdn_nam.n_name, host, sin->sdn_nam.n_len + 1);

    if (PROT_TRACE) fprintf(TDEST,  
	"DECnet: Parsed address as object number %d on host %.6s...\n",
		      sin->sdn_objnum, host);

#else /* Internet */

    /*	Parse host number if present */
    {
	BOOL numeric = YES;
	char *strptr = host;
	while (*strptr) {
	    if (!isdigit(*strptr) && *strptr != '.') {
		numeric = NO;
		break;
	    }
	    ++strptr;
	}
	if (numeric) {
	    sin->sin_addr.s_addr = inet_addr(host);	  /* See arpa/inet.h */
	} else {
	    if ((status = HTGetHostByName(host, sin, use_cur)) < 0) {
		free(host);
		return -1;
	    }
	}
	if (PROT_TRACE) {
	    fprintf(TDEST, "ParseInet... Parsed address as port %d on %s\n",
		    (int) ntohs(sin->sin_port),
		    HTInetString(sin));
	}
    }
#endif  /* Internet vs. Decnet */
    free(host);
    return status;
}


/* ------------------------------------------------------------------------- */
/*	       	      CONNECTION ESTABLISHMENT MANAGEMENT 		     */
/* ------------------------------------------------------------------------- */

/*								HTDoConnect()
**
**	Note: Any port indication in URL, e.g., as `host:port' overwrites
**	the default_port value.
**
**	Returns 0 if OK, -1 on error
*/
PUBLIC int HTDoConnect ARGS5(HTNetInfo *, net, char *, url,
			     u_short, default_port, u_long *, addr,
			     BOOL, use_cur)
{
  int status;
  char *p1 = HTParse(url, "", PARSE_HOST);
  char *at_sign;
  char *host;
  int flag;
  
  /* if theres an @ then use the stuff after it as a hostname */
  if((at_sign = strchr(p1, '@')) != NULL)
    host = at_sign+1;
  else
    host = p1;
  if (!*host) {
    HTErrorAdd(net->request, ERR_FATAL, NO,HTERR_NO_HOST,NULL,0,"HTDoConnect");
    free(p1);
    return -1;
  } else {
    if (PROT_TRACE) fprintf(TDEST, "HTDoConnect. Looking up `%s\'\n", host);
  }
  
  /* Set up defaults */
  if (net->sockfd==INVSOC) {
    memset((void *) &net->sock_addr, '\0', sizeof(net->sock_addr));
#ifdef DECNET
    net->sock_addr.sdn_family = AF_DECnet;/* Family = DECnet, host order */
    net->sock_addr.sdn_objnum = DNP_OBJ;  /* Default: http object number */
#else  /* Internet */
    net->sock_addr.sin_family = AF_INET;
    net->sock_addr.sin_port = htons(default_port);
#endif
  }
  
  /* If we are trying to connect to a multi-homed host then loop here until
     success or we have tried all IP-addresses */
  do {
    if (net->sockfd==INVSOC) {
      int hosts;
      if ((hosts = HTParseInet(&net->sock_addr, host, use_cur)) < 0) {
	if (PROT_TRACE)
	  fprintf(TDEST,"HTDoConnect. Can't locate remote host `%s\'\n", host);
	HTErrorAdd(net->request, ERR_FATAL, NO, HTERR_NO_REMOTE_HOST,
		   (void *) host, strlen(host), "HTDoConnect");
	break;
      };
      if (!net->addressCount && hosts > 1) net->addressCount = hosts;
#ifdef DECNET
      if ((net->sockfd=socket(AF_DECnet, SOCK_STREAM, 0))==INVSOC)
#else
	if ((net->sockfd=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP))==INVSOC)
#endif
	  {
	    HTErrorSysAdd(net->request, ERR_FATAL, socerrno, NO, "socket");
	    break;
	  }
      
      RoverSetCloseExec(net->sockfd);	/* Set to close on exec */

      if (RoverDebug) fprintf(stderr, ">>> OPEN %d\n", net->sockfd);
      flag = 1;
      if (setsockopt(net->sockfd, IPPROTO_TCP, TCP_NODELAY,
		     (char *) &flag, sizeof(int)) < 0)
	if (PROT_TRACE)
	  fprintf(TDEST,"HTDoConnect. Error setting TCP_NODELAY opt.\n");
      
      if (addr)	*addr = ntohl(net->sock_addr.sin_addr.s_addr);
      if (PROT_TRACE) fprintf(TDEST, "HTDoConnect. Created socket number %d\n",
			      net->sockfd);
      
      /* If non-blocking protocol then change socket status */
      if (!HTProtocolBlocking(net->request)) {
	status = RoverMakeSockNB(net->sockfd, 0);
	if (status == -1) {
	  if (PROT_TRACE)
	    fprintf(TDEST, "HTDoConnect. Can NOT make socket non-blocking\n");
	} else if (PROT_TRACE)
	  fprintf(TDEST, "HTDoConnect. Using NON_BLOCKING I/O\n");
      }
      
      /* If multi-homed host then start timer on connection */
      if (net->addressCount >= 1) net->connecttime = time(NULL);
    }
    
    /* Check for interrupt */
    if (HTThreadIntr(net->sockfd)) {
      fprintf(stderr, ">>>>>>>>>>>>>>>>>>>> Got interrupt in DoConnect\n");
      if (NETCLOSE(net->sockfd) < 0)
	HTErrorSysAdd(net->request, ERR_FATAL, socerrno,NO,"NETCLOSE");
      HTThreadState(net->sockfd, THD_CLOSE);
      net->sockfd = INVSOC;
      free(p1);
      return HT_INTERRUPTED;
    }
    
    /* Do a connect */
    status = connect(net->sockfd, (struct sockaddr *) &net->sock_addr,
		     sizeof(net->sock_addr));
    /*
     * According to the Sun man page for connect:
     *     EINPROGRESS         The socket is non-blocking and the  con-
     *                         nection cannot be completed immediately.
     *                         It is possible to select(2) for  comple-
     *                         tion  by  selecting the socket for writ-
     *                         ing.
     * According to the Motorola SVR4 man page for connect:
     *     EAGAIN              The socket is non-blocking and the  con-
     *                         nection cannot be completed immediately.
     *                         It is possible to select for  completion
     *                         by  selecting  the  socket  for writing.
     *                         However, this is only  possible  if  the
     *                         socket  STREAMS  module  is  the topmost
     *                         module on  the  protocol  stack  with  a
     *                         write  service  procedure.  This will be
     *                         the normal case.
     */
    if ((status < 0) && InProgress(socerrno)) {
      if (PROT_TRACE) fprintf(TDEST, "HTDoConnect. WOULD BLOCK `%s'\n", host);
      HTThreadState(net->sockfd, THD_SET_WRITE);
      free(p1);
      return HT_WOULD_BLOCK;
    }
    
    if (PROT_TRACE) fprintf(TDEST, "HTDoConnect. Connected to `%s' on %d\n", host, net->sockfd);
    flag = 1;
    if (setsockopt(net->sockfd, IPPROTO_TCP, TCP_NODELAY,
		   (char *) &flag, sizeof(int)) < 0)
      if (PROT_TRACE) 
	fprintf(TDEST,"HTDoConnect. Error setting TCP_NODELAY opt.\n");
    
#ifdef decstation
    if (!HTProtocolBlocking(net->request)) {
      status = RoverMakeSockNB(net->sockfd, 1);
      if (status == -1) {
	if (PROT_TRACE)
	  fprintf(stderr, "HTDoConnect. Can NOT make socket non-blocking\n");
      } else if (PROT_TRACE)
	fprintf(stderr, "HTDoConnect. Using NON_BLOCKING I/O\n");
    }
#endif /* decstation */

    if (net->addressCount >= 1) {
      net->connecttime = time(NULL) - net->connecttime;
      if (status < 0) {
	if (socerrno==EISCONN) {  /* connect multi after would block */
	  HTThreadState(net->sockfd, THD_CLR_WRITE);
	  TCPAddrWeights(host, net->connecttime);
	  free(p1);
	  net->addressCount = 0;
	  return 0;
	}
	HTErrorSysAdd(net->request, ERR_NON_FATAL, socerrno, NO, "connect");

	/* I have added EINVAL `invalid argument' as this is what I 
	   get back from a non-blocking connect where I should 
	   get `connection refused' on SVR4 */
	if (socerrno==ECONNREFUSED || socerrno==ETIMEDOUT ||
	    socerrno==ENETUNREACH || socerrno==EHOSTUNREACH ||
#ifdef __srv4__
	    socerrno==EINVAL ||
#endif
	    socerrno==EHOSTDOWN)
	  net->connecttime += TCP_DELAY;
	else
	  net->connecttime += TCP_PENALTY;
	if (NETCLOSE(net->sockfd) < 0)
	  HTErrorSysAdd(net->request, ERR_FATAL, socerrno, NO, "NETCLOSE");
	HTThreadState(net->sockfd, THD_CLOSE);
	net->sockfd = INVSOC;
	TCPAddrWeights(host, net->connecttime);
      } else {				   /* Connect on multi-homed */
	TCPAddrWeights(host, net->connecttime);
	free(p1);
	net->addressCount = 0;
	return 0;
      }
    } else if (status < 0) {
      if (socerrno==EISCONN) {     /* Connect single after would block */
	HTThreadState(net->sockfd, THD_CLR_WRITE);
	net->addressCount = 0;
	free(p1);
	return 0;
      } else {
	HTErrorSysAdd(net->request, ERR_FATAL, socerrno, NO, "connect");
	TCPCacheRemoveHost(host);
	if (NETCLOSE(net->sockfd) < 0)
	  HTErrorSysAdd(net->request, ERR_FATAL, socerrno, NO, "NETCLOSE");
	HTThreadState(net->sockfd, THD_CLOSE);
	break;
      }
    } else {				  /* Connect on single homed */
      free(p1);
      net->addressCount = 0;
      return 0;
    }
  } while (--net->addressCount);
  
  if (PROT_TRACE) fprintf(TDEST, "HTDoConnect. Connect failed\n");
  free (p1);
  net->addressCount = 0;
  net->sockfd = INVSOC;
  return -1;
}




