 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: RoverBuffer.c
 *
 * Description: Provides simple variable-length buffer management
 *
 *
 ***********************************************************************
 *   Copyright 1997
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "RoverBase.h"
#include "RoverBuffer.h"


/* Allocate and initialize a buffer object */
PUBLIC RoverBuffer *RoverBufferInit(int size)
{
  RoverBuffer *rv;

  if (size <= 0) {
    fprintf(stderr,
	    "Illegal initial buffer size %d [RoverBufferInit:RoverBuffer.c]\n",
	    size);
    return(NULL);
  }
  if ((rv = (RoverBuffer *)malloc(sizeof(RoverBuffer))) == NULL) {
    fprintf(stderr,"Out of memory (buffer) [RoverBufferInit:RoverBuffer.c]\n");
    return(NULL);
  }
  if ((rv->data = (void *)calloc(size, sizeof(char))) == NULL) {
    fprintf(stderr, "Out of memory (data) [RoverBufferInit:RoverBuffer.c]\n");
    free(rv);
    return(NULL);
  }
  rv->alloc = size;
  rv->size = 0;
  return(rv);
}

/* Returns the size of a buffer */
PUBLIC int RoverBufferSize(RoverBuffer *buf)
{
  return(buf->size);
}

/* Returns the buffer object's data buffer */
PUBLIC void *RoverBufferBuf(RoverBuffer *buf)
{
  return(buf->data);
}

/* Deallocate a buffer object and its data buffer */
PUBLIC void RoverBufferFree(RoverBuffer *buf)
{
  free(buf->data);
  free(buf);
}

/* Append binary data to a buffer object */
PUBLIC void RoverBufferAppendBin(RoverBuffer *buf, void *data, int size)
{
  int i, j = buf->size + size;
  if (j >= buf->alloc) {
    do {
      buf->alloc <<= 1;
      } while (j >= buf->alloc);
    buf->data = (void *)realloc(buf->data, buf->alloc);
    i = buf->size+size+1;
    memset((void *)buf->data+i, 0, buf->alloc-i);
  }
  memcpy(buf->data+buf->size, data, size);
  buf->size = j;
}

/* Append a string to a buffer object */
PUBLIC void RoverBufferAppend(RoverBuffer *buf, char *data)
{
  RoverBufferAppendBin(buf, data, strlen(data));
}

/* Append a character to a buffer object */
PUBLIC void RoverBufferAppendC(RoverBuffer *buf, char data)
{
  RoverBufferAppendBin(buf, &data, 1);
}

/* Append a NULL character to a buffer object */
PUBLIC void RoverBufferTerminate(RoverBuffer *buf)
{
  char c = '\0';
  RoverBufferAppendBin(buf, &c, 1);
}

/* Set a location within a buffer to a binary value */
PUBLIC void RoverBufferSetBin(RoverBuffer *buf, void *data, int size, int idx)
{
  if (RoverDebug)
    fprintf(stderr, "buf size is %d, idx is %d, data size is %d\n",
	    buf->size, idx, size);
  if (!((idx+size)<=buf->size)) {
    fprintf(stderr,
	    "Illegal idx & size (%d + %d > %d) [RoverBufferSetBin:RoverBuffer.c]\n",
	    idx, size, buf->size);
  }
  /*  ASSERT((idx+size)<=buf->size); */
  memcpy(buf->data+idx, data, size);
}


