 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: RoverEscape.c
 *
 * Description: Common server and client URL escape functions
 *
 *
 ***********************************************************************
 *   Copyright 1997
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#define HEX_ESCAPE	'%'
#define SPACE_ESCAPE	'+'
#define TCL_ESCAPE	'\\'
#define ACCEPTABLE(A, MASK) \
	( ((A)>=32) && ((A)<128) && ((isAcceptable[(A)-32]) & (MASK)))

#define OTHER_ACCEPTABLE(A, MASK) \
	( ((A)<32) || ((A)>=128) || ((isAcceptable[(A)-32]) & (MASK)))

#include <stdio.h>
#include <stdlib.h>
#include "tcp.h"
#include "RoverBase.h"
#include "RoverEscape.h"


PRIVATE unsigned char isAcceptable[96] =

/* Overencodes */
/*      Bit 0 (1)       URL_XALPHAS          -- see HTFile.h
**      Bit 1 (2)       URL_XPALPHAS         -- as URL_XALPHAS but with plus.
**      Bit 2 (4)       URL_PATH             -- as URL_XPALPHAS but with /
**      Bit 3 (8)       TCL_ESC              -- Tcl escapes (nothing else)
**      Bit 4 (16)      SH_ESC               -- Tcl escapes but with \x0A
*/

/* Change 0x2B from 15 to 12 */
/* Change 0x25 from  8 to 24 to 16 */
/* Change 0x20 from  8 to 24 */
/*0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F          0123456789ABCDEF */
{24, 8, 0, 8, 0,16, 0, 0, 0, 0, 7,12, 8,15,15,12,  /* 2x   !"#$%&'()*+,-./  */
 15,15,15,15,15,15,15,15,15,15, 8, 0, 0, 8, 0, 0,  /* 3x  0123456789:;<=>?  */
 15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,  /* 4x  @ABCDEFGHIJKLMNO  */
 15,15,15,15,15,15,15,15,15,15,15, 0,24, 0, 0,15,  /* 5X  PQRSTUVWXYZ[\]^_  */
  0,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,  /* 6x  `abcdefghijklmno  */
 15,15,15,15,15,15,15,15,15,15,15, 0,24, 0, 0, 0   /* 7X  pqrstuvwxyz{\}~ DEL*/
};

PRIVATE char *hex = "0123456789ABCDEF";

/* ------------------------------------------------------------------------- */

/*              Escape undesirable characters using %           RoverEscape()
**              -------------------------------------
**
**      This function takes a pointer to a string in which
**      some characters may be unacceptable unescaped.
**      It returns a string which has these characters
**      represented by a '%' character followed by two hex digits.
**
**      In the tradition of being conservative in what you do and liberal
**      in what you accept, we encode some characters which in fact are
**      allowed in URLs unencoded -- so DON'T use the table below for
**      parsing! 
**
**      Unlike RoverUnEscape(), this routine may return a malloced string.
**
*/
/* The string returned from here, if any, can be free'd by caller. */
PUBLIC char *RoverEscapeBin(char *str, int *size, unsigned char mask)
{
  char *p, *q, *result;
  int i, count, unacceptable = 0;

  for(p = str, i = 0; i < *size; i++, p++)
    if (!ACCEPTABLE((unsigned char)TOASCII(*p), mask)) unacceptable++;
  if (!unacceptable) return(str);
  
  count = *size + unacceptable + unacceptable;
  if (!(result = (char *) malloc(count + 1))) {
    fprintf(stderr, "Out of memory [RoverEscapeBin:RoverEscape.c]\n");
    return(NULL);
  }
  for (q = result, p = str, i = 0; i < *size; i++, p++) {
    unsigned char a = TOASCII(*p);

    if (!ACCEPTABLE(a, mask)) {
      if (a == ' ') 	       /* For legacy applications, use `+' for ` ' */
	*q++ = SPACE_ESCAPE;
      else {
	*q++ = HEX_ESCAPE;  	/* Means next chars are hex */
	*q++ = hex[a >> 4];
	*q++ = hex[a & 15];
      }
    } else
      *q++ = *p;
  }
  *q++ = 0;                   /* Terminate */
  *size = count;
  return(result);
}

/* The string returned from here, if any, can be free'd by caller. */
PUBLIC char *RoverEscape(char *str, unsigned char mask)
{
  int size = strlen(str);
  return(RoverEscapeBin(str, &size, mask));
}

/*              Decode %xx escaped characters                   RoverUnescape()
**              -----------------------------
**
**      This function takes a pointer to a string in which some
**      characters may have been encoded in %xy form, where xy is
**      the acsii hex code for character 16x+y. Maps anything illegal to 0.
**      The string is converted in place, as it will never grow.
*/

PRIVATE inline char from_hex(char c)
{
  return((c >= '0' && c <= '9') ? (c - '0')
	 : ((c >= 'A' && c <= 'F') ? (c - 'A' + 10)
	    /* accept small letters just in case */
            : ((c >= 'a' && c <= 'f') ? (c - 'a' + 10)
	       : 0)));
}

PUBLIC char *RoverUnescape(char *str)
{
  char *p = str;
  char *q = str;

  if (!str) {                                       /* Just for safety ;-) */
    if (RoverDebug)
      fprintf(stderr, "RoverUnescape.. Called with NULL argument.\n");
    return(str);
  }

  while (*p) {
    if (*p == HEX_ESCAPE) {
      p++;
      if (*p) *q = from_hex(*p++) << 4;
      if (*p) *q = FROMASCII(*q + from_hex(*p++));
      q++;
    } else if (*p == SPACE_ESCAPE) {/* For legacy apps, accept `+' as ` ' */
      *q++ = ' ';
      p++;
    } else {
      *q++ = *p++; 
    }
  }
    
  *q = '\0';
  return(str);
}

/*
 * Unescape any spaces (encoded as _) in str.
 */
PUBLIC void bartospace(char *str) {
  register int x;
  
  for(x = 0; str[x]; x++) if(str[x] == '_') str[x] = ' ';
}

PRIVATE char *RoverEscapeOthers(char *str, unsigned char mask)
{
  char *p, *q, *result;
  int unacceptable = 0;

  for(p = str; *p; p++)
    if (!OTHER_ACCEPTABLE((unsigned char)TOASCII(*p), mask)) unacceptable++;
  if (!unacceptable) return(str);

  if (!(result = (char *) malloc(p-str + unacceptable + 1))) {
    fprintf(stderr, "Out of memory [RoverEscapeTcl:RoverEscape.c]\n");
    return(NULL);
  }
  for (q = result, p = str; *p; p++) {
    if (!OTHER_ACCEPTABLE((unsigned char)TOASCII(*p),mask)) {
      *q++ = TCL_ESCAPE;	/* Means special Tcl char is next */
    }
    *q++ = *p;
  }
  *q++ = 0;                   /* Terminate */
  return(result);
}

/*
 * Escape a string for the purpose of using it within Tcl.
 *   Allocates a new string if any escaping occurs.
 */
PUBLIC char *RoverEscapeTcl(char *str)
{
  return(RoverEscapeOthers(str, TCL_ESC));
}

/*
 * Escape a string for the purpose of using it as a shell command.
 *   Allocates a new string if any escaping occurs.
 */
PUBLIC char *RoverEscapeShellCmd(char *str)
{
  return(RoverEscapeOthers(str, (TCL_ESC|SH_ESC)));
}

/*
 * Return the index of ch in string str.
 */
PUBLIC int ind(char *str, char ch)
{
  register int x;
  
  for (x = 0; str[x]; x++) if (str[x] == ch) return(x);
  return(-1);
}

