/*				    				   RoverEvent.c
**	MULTITHREADED ACCESS MANAGER
**
**	(c) COPYRIGHT CERN 1994.
**	Please first read the full copyright statement in the file COPYRIGH.
**
** Authors:
**	HFN    	Henrik Frystyk Nielsen <frystyk@dxcern.cern.ch>
**
** History:
**      05 Jul 95 Written from scratch
**
** Bugs
**
*/

/* Library include files */
#include "RoverBase.h"
#include "tcp.h"
#include "HTUtils.h"
#include "HTAccess.h"
#include "HTError.h"
#include "HTThread.h"
#include "HTList.h"
#include "RoverEvent.h"					 /* Implemented here */

PUBLIC HTList *RoverConversions = NULL;	       /* Format conversions */

/* Type definitions and global variables etc. local to this module */
PRIVATE HTList *UserReadList = NULL;	  /* List of read callback functions */
PRIVATE HTList *UserWriteList = NULL;	 /* List of write callback functions */
PRIVATE	fd_set	HTfd_user_read;	      /* Bit array of read sockets from user */
PRIVATE	fd_set	HTfd_user_write;     /* Bit array of write sockets from user */
PRIVATE HTEventHandler TimerCallback;	    /* Timer callback for event loop */
PRIVATE struct timeval SelectTimeout;	       	  	   /* Select timeout */

/* --------------------------------------------------------------------------*/

/*							        HTEventRegister
**
**	Registers a socket as active for type together with a callback
**	function to call when select returns that the socket is pending.
*/
PUBLIC BOOL HTEventRegister ARGS2(HTEventCallBack*, user_fd, HTEventType, type)
{
    static BOOL done=NO;
    if (!user_fd) {
	if (THD_TRACE)
	    fprintf(TDEST, "EventRegist. Bad argument\n");
	return NO;
    }
    if (!done) {
	FD_ZERO(&HTfd_user_read);
	FD_ZERO(&HTfd_user_write);
	done = YES;
    }
    switch (type) {
    case EVENT_READ:
      if (!UserReadList)
	UserReadList = HTList_new();
      HTList_addObject(UserReadList, (void *) user_fd);
      HTThreadState(user_fd->sockfd, THD_SET_READ);
      FD_SET(user_fd->sockfd, &HTfd_user_read);
      break;
    case EVENT_WRITE:
      if (!UserWriteList)
	UserWriteList = HTList_new();
      HTList_addObject(UserWriteList, (void *) user_fd);
      HTThreadState(user_fd->sockfd, THD_SET_WRITE);
      FD_SET(user_fd->sockfd, &HTfd_user_write);
      break;
    default:
      fprintf(TDEST, "EventRegist. Bad type\n");
      return NO;
      break;
    }
    return YES;
}


/*							      HTEventDeregister
**
**	Deregisters a socket as active for type.
*/
PUBLIC BOOL HTEventDeregister ARGS2(HTEventCallBack*, user_fd, HTEventType, type)
{
    if (!user_fd) {
	if (THD_TRACE)
	    fprintf(TDEST, "EventDeregist. Bad argument\n");
	return NO;
    }
    switch (type) {
    case EVENT_READ:
      if (!UserReadList) {
	if (THD_TRACE)
	  fprintf(TDEST, "EventDeregist. No events registered\n");
	return NO;
      }
      if (HTList_removeObject(UserReadList, (void *) user_fd) == NO) {
	if (THD_TRACE)
	  fprintf(TDEST, "EventDeregist. Callback not found\n");
	return NO;
      }
      HTThreadState(user_fd->sockfd, THD_CLR_READ);
      FD_CLR(user_fd->sockfd, &HTfd_user_read);
      break;
    case EVENT_WRITE:
      if (!UserWriteList) {
	if (THD_TRACE)
	  fprintf(TDEST, "EventDeregist. No events registered\n");
	return NO;
      }
      if (HTList_removeObject(UserWriteList, (void *) user_fd) == NO) {
	if (THD_TRACE)
	  fprintf(TDEST, "EventDeregist. Callback not found\n");
	return NO;
      }
      HTThreadState(user_fd->sockfd, THD_CLR_WRITE);
      FD_CLR(user_fd->sockfd, &HTfd_user_write);
      break;
    default:
      fprintf(TDEST, "EventDeegist. Bad type\n");
      return NO;
      break;
    }
    return YES;
}


/*							        HTTimerRegister
**
**	Registers a callback function to be called from within the event loop
**	each time through the loop or after a select has timed out.
*/
PUBLIC BOOL HTTimerRegister ARGS2(HTEventHandler, callback, int, timeout)
{
    if (callback) {
        if(TimerCallback && (TimerCallback != callback)) {
            fprintf(TDEST,"TimerRegist. Timer callback already registered\n");
	    return NO;
	} else {
	    TimerCallback = callback;
	}
        SelectTimeout.tv_sec = timeout;
        SelectTimeout.tv_usec = 0;
    } else {
        TimerCallback = NULL;
        SelectTimeout.tv_sec = 0;
	SelectTimeout.tv_usec = 0;
    }
    return YES;
}


/*							         HTEventCleanup
**
**	Clears the list of registered user sockets and call back functions
*/
PUBLIC void HTEventCleanup NOARGS
{
    if (UserReadList) {
	HTList_delete(UserReadList);
	UserReadList = NULL;
	FD_ZERO(&HTfd_user_read);
	FD_ZERO(&HTfd_user_write);
    }
    if (UserWriteList) {
	HTList_delete(UserWriteList);
	UserWriteList = NULL;
	FD_ZERO(&HTfd_user_read);
	FD_ZERO(&HTfd_user_write);
    }
}


#ifdef ROVER_AM
/*							      HTEventCheckState
**
**	This function checks the return code from the client call back 
**	function HTEventHandler() and HTEventRequestTerminate().
**
**	Returns:
**		YES	If we stay in eventloop
**		NO	If we return to client application
*/
PRIVATE BOOL HTEventCheckState ARGS2(HTRequest *, request, HTEventState, ret)
{
    switch (ret) {
      case EVENT_TERM:					   /* Ignore it here */
      case EVENT_OK:
	break;
	
      case EVENT_INTR:
 	{
	    int sockfd;
	    if (request && request->net_info &&
		(sockfd = request->net_info->sockfd) >= 0)
		HTThreadState(sockfd, THD_SET_INTR);
	}
	break;

      case EVENT_INTR_ALL:
	HTThreadMarkIntrAll(&HTfd_user_read);
	HTThreadMarkIntrAll(&HTfd_user_write);
	break;
	
	break;

      case EVENT_QUIT:
	return NO;
	break;	       
	
      default:
	if (THD_TRACE)
	    fprintf(TDEST, "HTEventLoop. Invalid return code\n");
    }
    return YES;
}    
#endif /* ROVER_AM */


/*							    HTEventLoopNull
**
**	This is the internal socket event loop that is executed when the
**	client doesn't want to keep the event loop. The client can interrupt
**	this loop by sending an event on on of the registered sock descriptors
**	in HTThreadInit() in the HTThread.c module
**
**
**	Returns:
**		HT_ERROR	if can't get first document or error in select
**				(only if it failed with no blocking)
**		HT_OK		On normal exit
*/
PUBLIC int HTEventLoopNull NOARGS
{
    BOOL user_event = NO;
    int cnt, retry = 0, selres;
    HTRequest *cur_request = NULL;
    struct timeval timeout;	       	  	   /* Select timeout */
    HTEventState state;
#ifdef ROVER_AM
    int status = HT_OK;
    HTErrSeverity severity;
#endif /* ROVER_AM */

    while (1) {
        fd_set fd_read, fd_write;		      /* Set up working copy */
	int maxfdpl = HTThreadGetFDInfo(&fd_read, &fd_write);

	if (THD_TRACE)
	  fprintf(TDEST, "HTEventLoopNull: Selecting now...\n");
	/* Why do we continually reset the timeout? Because of !#$@@ Linux */
        timeout.tv_sec = SelectTimeout.tv_sec;
        timeout.tv_usec = SelectTimeout.tv_usec;

#ifdef __hpux
	selres = select(maxfdpl, (int *) &fd_read, (int *) &fd_write,
			(int *) NULL, &timeout);
#else
	selres = select(maxfdpl, &fd_read, &fd_write, NULL, &timeout);
#endif
	if (THD_TRACE)
	    fprintf(TDEST, "HTEventLoop. Select returns: %d\n", selres);
	if (selres < 0) {
	    perror("Select error [HTEventLoopNull:HTEvent.c]");
	    if (++retry >= 3) {
	      return HT_ERROR;
	    } else continue;
	}
	retry = 0;
	if (!selres) {
	    if (TimerCallback) {
	        if (THD_TRACE)
		    fprintf(TDEST, "HTEventLoop. Calling timer handler\n");
	        state = (*TimerCallback)(&cur_request);
		if (state == EVENT_TERM) {
		  TimerCallback = NULL;
		  SelectTimeout.tv_sec = 0;
		  SelectTimeout.tv_usec = 0;
		} else if (state == EVENT_QUIT)
		  return(HT_OK);
	    }
#ifdef ROVER_AM
	    goto check_lib;	/* Check for interrupted library sockets */
#else
	    continue;
#endif /* ROVER_AM */
	}
	if (THD_TRACE) {
	    int cnt;
	    fprintf(TDEST, "HTEventLoop. Sockets ready for read : ");
	    for (cnt=0; cnt<maxfdpl; cnt++) {
		if (FD_ISSET(cnt, &fd_read))
		    fprintf(TDEST, "%d ", cnt);
	    }
	    fprintf(TDEST, "\n");
	    fprintf(TDEST, "HTEventLoop. Sockets ready for write: ");
	    for (cnt=0; cnt<maxfdpl; cnt++) {
		if (FD_ISSET(cnt, &fd_write))
		    fprintf(TDEST, "%d ", cnt);
	    }
	    fprintf(TDEST, "\n");
	}

        /* Any events on a user socket? - all sockets are examined */
	for (cnt=0; cnt<maxfdpl; cnt++) {
	    if (FD_ISSET(cnt, &fd_read) && FD_ISSET(cnt, &HTfd_user_read)) {
		HTList *cur = UserReadList;
		HTEventCallBack *pres, *next;
		pres = (HTEventCallBack *) HTList_nextObject(cur);
		while (pres) {
		    /* Grab the next object now, since the handler call
		       might deallocate the current object. That would
		       leave us with a possibly stale pointer
		     */
		    next = (HTEventCallBack *) HTList_nextObject(cur);
		    if ((pres->sockfd == cnt) && 
			FD_ISSET(cnt, &HTfd_user_read)) {
			if (THD_TRACE)
			    fprintf(TDEST, "HTEventLoop. Calling read handler 0x%x for %d\n",
				    pres->callback, cnt);
			state = (*(pres->callback))((HTRequest **) cnt);
			if (state == EVENT_TERM) {
			  if (RoverDebug)
			    fprintf(stderr, "HTEventLoop. Deregistering for read on %d\n", cnt);
			  HTEventDeregister(pres, EVENT_READ);
			} else if (state == EVENT_QUIT)
			  return(HT_OK);
			user_event = YES;
		    }
		    pres = next;
		}
	    }
	    if (FD_ISSET(cnt, &fd_write) && FD_ISSET(cnt, &HTfd_user_write)) {
		HTList *cur = UserWriteList;
		HTEventCallBack *pres, *next;
		pres = (HTEventCallBack *) HTList_nextObject(cur);
		while (pres) {
		    /* Grab the next object now, since the handler call
		       might deallocate the current object. That would
		       leave us with a possibly stale pointer
		     */
		    next = (HTEventCallBack *) HTList_nextObject(cur);
		    if ((pres->sockfd == cnt) &&
			FD_ISSET(cnt, &HTfd_user_write)) {
		        if (THD_TRACE)
			    fprintf(TDEST,
				    "HTEventLoop. Calling write handler 0x%p\n",
				    pres->callback);
			state = (*(pres->callback))((HTRequest **) cnt);
			if (state == EVENT_TERM) 
			  HTEventDeregister(pres, EVENT_WRITE);
			else if (state == EVENT_QUIT)
			  return(HT_OK);
			user_event = YES;
		    }
		    pres = next;
		}
	    }
	}

#ifdef ROVER_AM
	if (user_event) {
	    user_event = NO;
	    continue;			 /* Go check for any new User events */
	} else {
	    if (THD_TRACE)
		fprintf(TDEST, "HTEventLoop. No user events found - look for Library events\n");
	}

      check_lib:
	/* Any events on a Library socket? - only do this when no user sockets
	   are pending */
	if ((cur_request = HTThread_getRequest(&fd_read, &fd_write)) != NULL) {
	  /* The Library call-back function should be decided in HTProtocol
	     structure as a function of the method used */

	  HTProtocol *callback = (HTProtocol *)
	    HTAnchor_protocol(cur_request->anchor);
	  status = (*callback->load)(cur_request);
	  if (THD_TRACE)
	    fprintf(TDEST, "HTEventLoop. Load status: %d\n", status);
	  if (status != HT_WOULD_BLOCK) {
	    /* The error stack might contain general information to the
	       client about what has been going on in the library (not
	       only errors) */
	    if (cur_request->error_stack) RoverErrorMsg(cur_request,
							&severity);
/*	    HTLoadTerminate(cur_request, status); */
	    if (THD_TRACE) fprintf(TDEST,"HTEventLoop. Calling Terminate\n");
	    HTErrorFree(cur_request);
	    state = HTEventRequestTerminate(cur_request, status);
	    if (!HTEventCheckState(cur_request, state)) {
	      return HT_OK;
	    }
	  }
	} else {
	    if (THD_TRACE)
		fprintf(TDEST, "HTEventLoop. No library events found\n");
	}
#endif /* ROVER_AM */
    }
    return HT_ERROR;				    /* Should never get here */
}


