 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTRoverTCP.c
 *
 * Description: Common server and client TCP and other functions
 *
 *
 ***********************************************************************
 *   Copyright 1997
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>
#include <netdb.h>
#include "RoverBase.h"
#include "tcp.h"
#include "HTUtils.h"
#include "HTString.h"
#include "RoverHost.h"

#ifndef RESOLV_CONF
#define RESOLV_CONF "/etc/resolv.conf"
#endif

PUBLIC char *RoverHostName;
PUBLIC char *RoverUserName;
PUBLIC char *RoverMailAddress;

PUBLIC void str_tolower(char *str)
{
  while(*str) {
    *str = tolower(*str);
    ++str;
  }
}
        
/*							     RoverGetDomainName
**	Returns the current domain name without the local host name.
**	The response is pointing to a static area that might be changed
**	using RoverSetHostName().
**
**	Returns NULL on error, "" if domain name is not found
*/
PUBLIC char *RoverGetDomainName()
{
  char *host = RoverGetHostName();
  char *domain;

  if (host && *host) {
    if ((domain = strchr(host, '.')) != NULL)
      return(++domain);
    else
      return("");
  } else
    return(NULL);
}



/*							       RoverGetHostName
**	Returns the name of this host. It uses the following algorithm:
**
**	1) gethostname()
**	2) if the hostname doesn't contain any '.' try to read
**	   /etc/resolv.conf. If there is no domain line in this file then
**	3) Try getdomainname and do as the man pages say for resolv.conf (sun)
**		If there is no domain line in this file, then it is derived
**		from the domain name set by the domainname(1) command, usually
**		by removing the first component. For example, if the domain-
**		name is set to ``foo.podunk.edu'' then the default domain name
**		used will be ``pudunk.edu''.
**
**	This is the same procedure as used by res_init() and sendmail.
**
**	Return: hostname on success else NULL
*/
PUBLIC char *RoverGetHostName()
{
  int got_it = 0;
  FILE *fp;
  char name[MAXHOSTNAMELEN+1];

  if (RoverHostName) {	       	/* If already done */
    if (*RoverHostName)
      return(RoverHostName);
    else
      return(NULL);		/* We couldn't get the last time */
  }
  *(name+MAXHOSTNAMELEN) = '\0';
  if (gethostname(name, MAXHOSTNAMELEN)) { 	     /* Maybe without domain */
    if (RoverDebug)
      fprintf(stderr,"Error getting hostname [RoverGetHostName:RoverTCP.c]\n");
    return(NULL);
  }
  if (RoverDebug)
    fprintf(stderr, "Local host name is `%s\' [RoverGetHostName:RoverTCP.c]\n",
	    name);
  StrAllocCopy(RoverHostName, name);
  {
    char *strptr = strchr(RoverHostName, '.');
    if (strptr != NULL)				   /* We have it all */
      got_it = YES;
  }

#ifndef VMS
  /* Now try the resolver config file */
  if (!got_it && (fp = fopen(RESOLV_CONF, "r")) != NULL) {
    char buffer[80];
    *(buffer+79) = '\0';
    while (fgets(buffer, 79, fp)) {
      if (!strncasecmp(buffer, "domain", 6)) {	
	char *domainstr = buffer+6;
	char *end;
	while (*domainstr == ' ' || *domainstr == '\t')
	  domainstr++;
	end = domainstr;
	while (*end && !isspace(*end))
	  end++;
	*end = '\0';
	if (*domainstr) {
	  StrAllocCat(RoverHostName, ".");
	  StrAllocCat(RoverHostName, domainstr);
	  got_it = YES;
	  break;
	}
      }
    }
    fclose(fp);
  }
  
  /* If everything else has failed then try getdomainname */
#ifndef NO_GETDOMAINNAME
  if (!got_it) {
    if (getdomainname(name, MAXHOSTNAMELEN)) {
      if (RoverDebug) 
	fprintf(stderr,
		"Can't get domain name [RoverGetHostName:RoverTCP.c]\n");
      StrAllocCopy(RoverHostName, "");
      return(NULL);
    }

    /* If the host name and the first part of the domain name are different
       then use the former as it is more exact (I guess) */
    if (strncmp(name, RoverHostName, (int) strlen(RoverHostName))) {
      char *domain = strchr(name, '.');
      if (domain) StrAllocCat(RoverHostName, domain);
    }
  }
#endif /* NO_GETDOMAINNAME */
#endif /* not VMS */

  str_tolower(RoverHostName);
  if (*(RoverHostName+strlen(RoverHostName)-1) == '.')
    /* Remove trailing dot */
    *(RoverHostName+strlen(RoverHostName)-1) = '\0';
  if (RoverDebug)
    fprintf(stderr, "Full host name is `%s\' [RoverGetHostName:RoverTCP.c]\n",
	    RoverHostName);
  return(RoverHostName);
}

PUBLIC char *RoverGetUserName()
{
  char *name;

  if (RoverUserName) {	       	/* If already done */
    if (*RoverUserName)
      return(RoverUserName);
    else
      return(NULL);		/* We couldn't get the last time */
  }
  if ((name = getpwuid(getuid())->pw_name) == NULL) {
    fprintf(stderr,
	    "Unable to determine user name [RoverGetUserName:RoverTCP.c]\n");
    StrAllocCopy(RoverUserName, "");
  } else 
    StrAllocCopy(RoverUserName, name);
  return(RoverUserName);
}

/*								HTSetHostName
**	Sets the current hostname inclusive domain name.
**	If this is not set then the default approach is used using
**	HTGetHostname().
*/
PUBLIC void RoverSetHostName(char *host)
{
  char *strptr;

  if (host && *host) {
    StrAllocCopy(RoverHostName, host);
    strptr = RoverHostName;
    str_tolower(strptr);
    if (*(RoverHostName+strlen(RoverHostName)-1) == '.')
      /* Remove trailing dot */
      *(RoverHostName+strlen(RoverHostName)-1) = '\0';
  } else {
    if (RoverDebug)
      fprintf(stderr, "Bad argument ignored [RoverSetHostName:RoverTCP.c]\n");
  }
}


/*
**	Free the host name. Called from HTLibTerminate
*/
PUBLIC void RoverFreeHostName()
{
  FREE(RoverHostName);
}

/*							    RoverSetMailAddress
**	Sets the current mail address.
**	This might be set by the user to change the value in the
**	From field in the HTTP Protocol.
**	If this is not set then the default approach is to use
**	RoverGetMailAddress(). If the argument is NULL or "" then 
**	RoverGetMailAddress returns NULL on a succeding request.
*/
PUBLIC void RoverSetMailAddress(char *address)
{
  if (!address || !*address)
    StrAllocCopy(RoverMailAddress, "");
  else
    StrAllocCopy(RoverMailAddress, address);
  if (RoverDebug)
    fprintf(stderr,
	    "Set mail address to `%s\' [RoverSetMailAddress:RoverHost.c]\n",
	    RoverMailAddress);
}




/*							    RoverGetMailAddress
**
**	This functions returns a char pointer to a static location
**	containing the mail address of the current user on the current
**	host. The static location is different from the one of the
**	current host name so different values can be assigned. The
**	default value is <LOGIN>@hostname where hostname is the one
**	initialized in RoverSetHostName or RoverGetHostName. The login
**	name is determined using (ordered):
**
**		getlogin
**		getpwuid(getuid())
**
**	The weakness about the last attempt is if the user has multiple
**	login names each with the same user ID. If this fails as well then:
**
**		LOGNAME environment variable
**		USER environment variable
**
**	Returns NULL if error else pointer to static string
*/
PUBLIC char *RoverGetMailAddress()
{
  char *login;
  char *domain;
  struct passwd *pw_info;

  if (RoverMailAddress) {
    if (*RoverMailAddress)
      return(RoverMailAddress);
    else
      return(NULL);       /* No luck the last time so we won't try again */
  }

#ifdef VMS
  if ((login = (char *) cuserid(NULL)) == NULL) {
    if (RoverDebug)
      fprintf(stderr, "cuserid returns NULL [RoverGetMailAddr:RoverHost.c]\n");
  }

#else
#ifdef _WINDOWS
  login = "PCUSER";				  /* @@@ COULD BE BETTER @@@ */
#else /* Unix like... */
  if ((login = (char *) getlogin()) == NULL) {
    if (RoverDebug)
      fprintf(stderr,"getlogin returns NULL [RoverGetMailAddr:RoverHost.c]\n");
    if ((pw_info = getpwuid(getuid())) == NULL) {
      if (RoverDebug)
	fprintf(stderr,
		"getpwid returns NULL [RoverGetMailAddr:RoverHost.c]\n");
      if ((login = getenv("LOGNAME")) == NULL) {
	if (RoverDebug)
	  fprintf(stderr,"LOGNAME not found [RoverGetMailAddr:RoverHost.c]\n");
	if ((login = getenv("USER")) == NULL) {
	  if (RoverDebug)
	    fprintf(stderr,"USER not found [RoverGetMailAddr:RoverHost.c]\n");
	  return(NULL);		/* I GIVE UP */
	}
      }
    } else
      login = pw_info->pw_name;
  }
#endif /* _WINDOWS */
#endif /* VMS */

  if (login) {
    StrAllocCopy(RoverMailAddress, login);
    StrAllocCat(RoverMailAddress, "@");
    if ((domain = RoverGetHostName()) != NULL)
      StrAllocCat(RoverMailAddress, domain);
    else {
      *RoverMailAddress = '\0';
      return(NULL);			/* Domain name not available */
    }
    return(RoverMailAddress);
  }
  return(NULL);
}


/*
**	Free the mail address. Called from HTLibTerminate
*/
PUBLIC void RoverFreeMailAddress()
{
  FREE(RoverMailAddress);
}


