 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: HTRoverTCP.c
 *
 * Description: Common server and client TCP and other functions
 *
 *
 ***********************************************************************
 *   Copyright 1997
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#define ROVER_ACCESSMGR

/* Define u_char, u_short, u_int, and u_long. */
#include <sys/types.h>

#include <sys/socket.h>
#include <netinet/tcp.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <strings.h>
#include "RoverBase.h"
#include "html.h"
#include "RoverTCP.h"
#include "tcp.h"
#include "HTUtils.h"
#include "RoverHost.h"

PRIVATE char *Canon(char **filename, char *host);

/* Breaks a URL up into its protocol, site name, port number, and path name
   (the path name also contains any query or anchor data). If the URL is
   relative, it is reformatted to be absolute, based on the URL given in
   baseurl. All strings produced (except error messages) are fresh copies 
   allocated from the heap. Returns NULL on success; otherwise an error
   message. */
PUBLIC char *parseURL(char **urlp, char **protocol, char **hostname,
		      int *port, char **urlpath, char *baseurl)
{
  int z;
  char *url = *urlp;
  char *p = url;

  *protocol = NULL;
  *hostname = NULL;
  *port = 80;
  *urlpath = NULL;

  for (z = 0; 1; z++) {
    if (p[z] == ':') break;
    if (p[z] == '/' || p[z] == '#' || p[z] == 0) { /* no protocol specified */
      char *s, *t;

      if (!baseurl) return "Bad URL: no protocol separator colon.";

      /* Get the protocol from the base url (add 1 byte in case of missing /)*/
      t = (char *) malloc(strlen(baseurl)+strlen(url)+2);
      strcpy(t, baseurl);
      if (*url == '/')
	s = strchr(strchr(strchr(t, '/')+1, '/')+1, '/');
      else {
	if (!(s = strrchr(t, '/')+1)) return "Bad base URL.";
	if (s == (strchr(t, '/')+2)) {
	  /* missing the trailing / */
	  strcat(t, "/");
	  s = t + strlen(t);
	}
      }
      strcpy(s, url);
      *urlp = t;
      return parseURL(urlp, protocol, hostname, port, urlpath, NULL);
    }
  }
  *protocol = newncopy(p, z);
  p = &p[z+1];
  if (*p++ !='/' || *p++ !='/') return "Bad URL: no :// after protocol.";
  for (z = 0; 1; z++) {
    if (p[z] == ':') {
      *hostname = newncopy(p, z);
      p = &p[z+1];
      *port = atoi(p);
      p = &p[strspn(p, "0123456789")];
      if (*p == 0) *urlpath = "/";
      else if (*p == '/') *urlpath = p;
      else return "Bad URL: bad pathname after port number.";
      break;
    }
    if (p[z] == '/') {
      *hostname = newncopy(p, z);
      *urlpath = &p[z];
      break;
    }
    if (!p[z]) {
      *hostname = newncopy(p, z);
      *urlpath = "/";
      break;
    }
  }    
  return NULL;
}

/*	        Simplify a URI
//		--------------
// A URI is allowed to contain the seqeunce xxx/../ which may be
// replaced by "" , and the seqeunce "/./" which may be replaced by "/".
// Simplification helps us recognize duplicate URIs. 
//
//	Thus, 	/etc/junk/../fred 	becomes	/etc/fred
//		/etc/junk/./fred	becomes	/etc/junk/fred
//

>> Now we eliminate leading ../ sequences (since the Netscape does) <<
  //      but we should NOT change
  //		http://fred.xxx.edu/../..
  //
  //	or	../../albert.html

//
// In the same manner, the following prefixed are preserved:
//
//	./<etc>
//	//<etc>
//
// In order to avoid empty URLs the following URLs become:
//
//		/fred/..		becomes /fred/..
//		/fred/././..		becomes /fred/..
//		/fred/.././junk/.././	becomes /fred/..
//
// If more than one set of `://' is found (several proxies in cascade) then
// only the part after the last `://' is simplified.
//

>> Now we also eliminate any hashes in the path<<
  //		/index.html#foo		becomes /index.html

// Returns: A string which might be the old one or a new one.
*/
PUBLIC char *Simplify(char *filename)
{
  char *path;
  char *p;
  
  if (!filename) {
    fprintf(stderr, "Simplify.. Bad argument\n");
    return filename;
  }

  if (RoverDebug) fprintf(stderr, "Simplify.. `%s\'\n", filename);

  if ((path = strstr(filename, "://")) != NULL) {	   /* Find host name */
    char *newptr;
    path += 3;
    while ((newptr = strstr(path, "://")) != NULL)
      path = newptr+3;
    path = Canon(&filename, path);       	      /* We have a host name */
  } else if ((path = strstr(filename, ":/")) != NULL) {
    path += 2;
  } else
    path = filename;
  if (*path == '/' && *(path+1)=='/') {	  /* Some URLs start //<foo> */
    path += 1;
  } else if (!strncmp(path, "news:", 5)) {
    char *ptr = strchr(path+5, '@');
    if (!ptr) ptr = path+5;
    str_tolower(ptr);			    /* Make group or host lower case */

/* fprintf(stderr, "into\n............ `%s'\n", filename); */
    return filename;		      /* Doesn't need to do any more */
  }
  if ((p = path)) {
    int segments = 0;
    
    /* Parse string first time to find number of `real' tokens */
    while (*p && (*p != '#')) {
      if (*p=='/' || p==path) {
	if (!((*(p+1)=='/' || !*(p+1)) ||
	      (*(p+1)=='.' && (*(p+2)=='/' || !*(p+2))) ||
	      (*(p+1)=='.' && *(p+2)=='.' &&(*(p+3)=='/' || !*(p+3)))))
	  segments++;
      }
      p++;
    }
    if (*p == '#') *p = '\0';	/* Eliminate hash and anything after it */

    /* Parse string second time to simplify */
    p = path;
/* fprintf(stderr, "Path is %s [%d]\n", path, segments); */
    while(*p) {
      if (*p=='/') {
	if (p>path && *(p+1)=='.' && (*(p+2)=='/' || !*(p+2))) {
	  char *orig=p, *dest=p+2;
	  while ((*orig++ = *dest++)); /* Remove a slash and a dot */
	  p--;
	} else if (/* segments>1 && */ *(p+1)=='.' && *(p+2)=='.' &&
		   (*(p+3)=='/' || !*(p+3))) {
	  char *q = p;
	  while (q>path && *--q!='/');	       /* prev slash */
/* fprintf(stderr, "q is %s\n", q); */
/*	  if (strncmp(q, "/../", 4) && strncmp(q, "/./", 3) &&
	      strncmp(q, "./", 2)) */
	    {
	    char *orig=q, *dest=p+3;
	    if (*q!='/') dest++;
	    while ((*orig++ = *dest++));	   /* Remove /xxx/.. */
	    segments--;
	    p = q-1;	      /* Start again with prev slash */
	  } /* else p++; */
	} else if (*(p+1)=='/') {
	  while (*(p+1)=='/') {
	    char *orig=p, *dest=p+1;
	    while ((*orig++ = *dest++));  /* Remove multiple /'s */
	  }
	}
      }
      p++;
    }  /* end while (*p) */
    if (p == path) {	/* Make sure we have a trailing / */
     if (!*p) { /* Need to extend by one */
      int len = strlen(filename);
      p = (char *) malloc(len+2);
      strcpy(p, filename);
      p += len;
     }
     *p = '/';
     *(p+1) = '\0';
   }
  }
/* fprintf(stderr, "into\n............ `%s'\n", filename); */
  return filename;
}

/*							       Canon
**
**	Canonicalizes the URL in the following manner starting from the host
**	pointer:
**
**	1) The host name is converted to lowercase
**	2) Expands the host name of the URL from a local name to a full
**	   domain name. A host name is started by `://'.
**	3) Chop off port if `:80' (http), `:70' (gopher), or `:21' (ftp)
**
**	Return: OK	The position of the current path part of the URL
**			which might be the old one or a new one.
*/
PRIVATE char *Canon(char **filename, char *host)
{
  char *newname = NULL;
  char *port;
  char *strptr;
  char *path;
  char *access = host-3;
  
  while (access>*filename && *(access-1)!='/')         /* Find access method */
    access--;
  if ((path = strchr(host, '/')) == NULL)			/* Find path */
    path = host + strlen(host);
  if ((strptr = strchr(host, '@')) != NULL && strptr<path)	   /* UserId */
    host = strptr;
  if ((port = strchr(host, ':')) != NULL && port>path)        /* Port number */
    port = NULL;
  
  strptr = host;				    /* Convert to lower-case */
  while (strptr<path) {
    *strptr = tolower(*strptr);
    strptr++;
  }
  
#if 0
    /* Does the URL contain a full domain name? This also works for a
       numerical host name. The domain name is already made lower-case
       and without a trailing dot. */
    if (((strptr = strchr(host, '.')) == NULL || strptr >= path) &&
	strncasecmp(host, "localhost", 9)) {
	CONST char *domain = RoverGetDomainName();
	if (domain) {
	    if ((newname = (char *) calloc(1, strlen(*filename) +
				       strlen(domain)+2)) == NULL)
		outofmem(__FILE__, "HTCanon");
	    if (port)
		strncpy(newname, *filename, (int) (port-*filename));
	    else
		strncpy(newname, *filename, (int) (path-*filename));
	    strcat(newname, ".");
	    strcat(newname, domain);
	}
    } else {					  /* Look for a trailing dot */
	char *dot = port ? port : path;
	if (dot > *filename && *--dot=='.') {
	    char *orig=dot, *dest=dot+1;
	    while((*orig++ = *dest++));
	    if (port) port--;
	    path--;
	}
    }
#endif

  /* Chop off port if `:80' (http), `:70' (gopher), or `:21' (ftp) */
  if (port) {
    if ((!strncmp(access, "http", 4) &&
	 (*(port+1)=='8'&&*(port+2)=='0'&&(*(port+3)=='/'||!*(port+3)))) ||
	(!strncmp(access, "gopher", 6) &&
	 (*(port+1)=='7'&&*(port+2)=='0'&&(*(port+3)=='/'||!*(port+3)))) ||
	(!strncmp(access, "ftp", 3) &&
	 (*(port+1)=='2'&&*(port+2)=='1'&&(*(port+3)=='/'||!*(port+3))))) {
      if (!newname) {
	char *orig=port, *dest=port+3;
	while((*orig++ = *dest++));
      }
    } else if (newname)
      strncat(newname, port, (int) (path-port));
  }
  
  if (newname) {
    char *newpath = newname+strlen(newname);
    strcat(newname, path);
    path = newpath;
    free(*filename);				    /* Free old copy */
    *filename = newname;
  }
  return path;
}

PUBLIC char *convertURL(URLinfo *uinfo)
{
  /* Ignore errors from the first parse */
  char *es = parseURL(&(uinfo->url), &(uinfo->protocol), &(uinfo->hostname),
		      &(uinfo->port), &(uinfo->urlpath), uinfo->base);
  uinfo->url = Simplify(strdup(uinfo->url));
  es = parseURL(&(uinfo->url), &(uinfo->protocol), &(uinfo->hostname),
		&(uinfo->port), &(uinfo->urlpath), uinfo->base);
  return(es);
}

/* Returns freshly malloc'd copy of partial string. Zero-terminates it, too. */
PUBLIC char *newncopy(char *s, int n)
{
  char *d = (char *) malloc(n+1);
  strncpy(d, s, n);
  d[n] = (char) 0;
  return d;
}

PUBLIC int RoverTCPOpenSocket(int port, int execClose)
{
  int roverSocket, flag;
  struct sockaddr_in serv_addr;

  /*
   * Open a TCP socket (an Internet stream socket).
   */
  if ((roverSocket = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    perror("Socket open failed [RoverTCPOpenSocket:RoverTCP.c]");
    return(-1);
  }

  if (execClose) RoverSetCloseExec(roverSocket);    /* Set to close on exec */

  flag = 1;
  if (setsockopt(roverSocket, IPPROTO_TCP, TCP_NODELAY, (char *)&flag,
		 sizeof(int)) < 0) {
    perror("TCP_NODELAY setsockopt error [RoverTCPOpenSocket:RoverTCP.c]");
    return(-1);
  }
  flag = 1;
  if (setsockopt(roverSocket, SOL_SOCKET, SO_REUSEADDR, (char*)&flag,
		 sizeof(int)) < 0){
    perror("SO_REUSEADDR setsockopt error [RoverTCPOpenSocket:RoverTCP.c]");
    return(-1);
  }
  
  /* From the NCSA httpd server code:
     Added at the suggestion of SGI, seems some PC clients don't close the 
     connection, which causes IRIX and Solaris to keep sending them packets
     for a really long time . . . 
     Sent in From: Brian Behlendorf <brian@wired.com> */
  flag = 1;
  if (setsockopt(roverSocket, SOL_SOCKET, SO_KEEPALIVE, (char*)&flag,
		sizeof(int)) < 0){
    perror("SO_KEEPALIVE setsockopt error [RoverTCPOpenSocket:RoverTCP.c]"); 
    return(-1); 
  }

  /*
   * Bind our local address so that others can send to us.
   */
  memset((void *) &serv_addr, 0, sizeof(serv_addr));
  serv_addr.sin_family      = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port        = htons(port);
  if (bind(roverSocket,(struct sockaddr *) &serv_addr,sizeof(serv_addr)) < 0) {
    perror("Local address bind failed [RoverTCPOpenSocket:RoverTCP.c]");
    fprintf(stderr, "Bind of local address to port %d failed\n", port);
    return(-1);
  }
    
  listen(roverSocket, 5);
  return(roverSocket);
}

/*
 * Read a line from a descriptor.  Read the line one byte at a time,
 * looking for the newline.  We store the newline in the buffer,
 * then follow it with a null (the same as fgets(3)).
 * Returns:
 *	GOT_NL		 	if a newline was found
 *	GOT_NONL		if no newline was found
 *	GOT_ERROR		if an error occurred (not EINTR or EAGAIN)
 *	GOT_EOF			if EOF occurred
 *	GOT_BUFFER_OVERFLOW	if buffer overflow occurred
 */
PUBLIC NBR readline(int fd, char *ptr, int *offset, int maxlen)
{
  int rc;
  char c;

  maxlen--;	/* Save a space for the ending null */
  ptr += *offset;
  for (; *offset < maxlen; (*offset)++) {
  idle_retry:
    if ((rc = read(fd, &c, 1)) == 1) {
      *ptr++ = c;
      if (c == '\n') {
	*ptr = '\0';
	return(GOT_NL);
      }
    } else if (rc == 0) {		/* EOF */
      return(GOT_EOF);
    } else if (rc == -1) {		/* Other error */
      rc = errno;
      if (RoverDebug) fprintf(stderr,
			      "readline err %d (%s) [readline:RoverTCP.c]\n",
			      rc, strerror(rc));
      if (WouldBlock(rc)) {
	return(GOT_NONL);		/* No pending data, we're done */
      } else if (rc == EINTR) {
	errno = 0;
	goto idle_retry;		/* Probably a profiling timer */
      } else {				/* Error */
	return(GOT_ERROR);
      }
    }
  }
  return(GOT_BUFFER_OVERFLOW);		/* Buffer overflow */
}

PUBLIC int readbuffer(int fd, char *ptr, int maxlen)
{
  int rc;

  maxlen--;				/* Save a space for the ending null */
  if (maxlen <= 0) return(-2);
 read_retry:
  if ((rc = read(fd, ptr, maxlen)) == 0) {	/* EOF */
      return(-1);
  } else if (rc == -1) {		/* Other error */
    rc = errno;
    if (RoverDebug) fprintf(stderr,
			    "readbuffer err %d (%s) [readbuffer:RoverTCP.c]\n",
			    rc, strerror(rc));
    if (WouldBlock(rc)) {
      return(0);			/* No pending data, we're done */
    } else if (rc == EINTR) {
      errno = 0;
      goto read_retry;			/* Probably a profiling timer */
    } else {				/* Error */
      return(-2);
    }
  }
  ptr[rc] = '\0';
  return(rc);
}

/*
 * Write a buffer to a descriptor. 
 * Returns:
 *	PUT_DONE	 	if the buffer is completely written
 *	PUT_CONTINUE		if more data remains to be sent
 *	PUT_ERROR		if an error occurred (not EINTR or EAGAIN)
 *	PUT_EOF			if EOF occurred
 */
PUBLIC NBW writebuffer(int fd, char *ptr, int *offset, int *size)
{
  int rc;

  ptr += *offset;
write_retry:
  rc = write(fd, ptr, *size);
  if (rc == 0) {		/* EOF */
    return(PUT_EOF);
  } else if (rc == -1) {	/* Other error */
    rc = errno;
    if (RoverDebug) fprintf(stderr,
			    "writebuffer err %d (%s) [writeuffer:RoverTCP.c]\n",
			    rc, strerror(rc));
    if (WouldBlock(rc)) {
      return(PUT_CONTINUE);
    } else if (rc == EINTR) {
      errno = 0;
      goto write_retry;		/* probably a profiling timer */
    } else {
      return(PUT_ERROR);
    }
  }
  
  /* Partial or complete write */
  if (RoverDebug)
    fprintf(stderr,"Wrote %d bytes to %d [writebuffer:RoverTCP.c]\n",rc,fd);
  *size -= rc;
  *offset += rc;
  if (!*size) return(PUT_DONE);
  else return(PUT_CONTINUE);
}

/*
 * Takes a buffer (ptr) and its size (size) and invokes fcn on each 
 *   NL-separated string in the buffer. Copies remaining data to the beginning
 *   of the buffer. Sets http->state to nextState when the fcn indicates
 *   it is done (by returning 1). Returns the number of bytes consumed from
 *   buffer. Otherwise, returns -1.
 */
PUBLIC int NLProcessLine(void *ClientData, NLFunction fcn, char *ptr, int size,
			 void *nextState)
{
  char *start = ptr, *nl, save, end;
  int rc = 0, diff, rem;

  end = *(ptr+size);
  *(ptr+size) = '\0';	/* Prevent running off the end */

  if (RoverDebug) fprintf(stderr, "NLPL got `%s'\n", ptr);

  while ((nl = index(ptr, '\n')) != NULL) {
    save = *(nl+1);
    *(nl+1) = '\0';
    if (RoverDebug) fprintf(stderr, "Calling NL function with [0x%x] `%s'\n",
			    (unsigned int) nl, ptr);
    rc = fcn(ClientData, ptr, nextState);
    *(nl+1) = save;
    ptr = nl+1;
    if (rc == -1) {
      fprintf(stderr, "NLFunction error [NLProcessLine:RoverTCP.c]\n");
      *(ptr+size) = end;
      fcn(ClientData, NULL, (void *) -1);
      return(-1);
    } else if (rc == 1)	{	/* Done processing NL delimited fields */
      fcn(ClientData, NULL, nextState);
      break;
    }
  }
  diff = ptr - start;
  rem = size - diff;
  if (RoverDebug) fprintf(stderr, "Size was %d, now is %d\n", size, rem);
  *(ptr+size) = end;
  
  if (rem) memmove((void *)start, (void *)ptr, rem); /* Save leftover data */
  return(diff);
}

/* Gets a line in a non-blocking manner:
   Returns 1 when the state has changed.
   Returns 0 if more data is needed.
 */
PUBLIC int TCPGetLine(NBTcp *tcp, StateChangeFunc func, void *ClientData,
		      void *nextState, void *abortState)
{
  int rc = readline(tcp->sock, tcp->line, &(tcp->offset), BUFSIZE);

  switch (rc) {
  case GOT_NL:	/* Newline found */
    (*func)(ClientData, nextState);
    break;
  case GOT_NONL:	/* No newline found, keep going */
    return(0);
    break;
  case GOT_ERROR:	/* An error occurred (other than EINTR or EAGAIN) */
    (*func)(ClientData, abortState);
    break;
  case GOT_EOF:	/* EOF occurred */
    fprintf(stderr, "Readline EOF [TCPGetLine:RoverTCP.c]\n");
    (*func)(ClientData, abortState);
    break;
  case GOT_BUFFER_OVERFLOW:
    fprintf(stderr, "Buffer overflow [TCPGetLine:RoverTCP.c]\n");
    (*func)(ClientData, abortState);
    break;
  }
  return(1);	/* State changed */
}

/* Gets a line in a non-blocking manner:
   Returns 1 when the state is changed.
   Returns 0 if more data is needed.
 */
PUBLIC int TCPRead(NBTcp *tcp, char *buf, int bufSize, int compress,
		   StateChangeFunc func, void *ClientData, void *nextState,
		   void *doneState, void *abortState)
{
  int rc;

 read_retry:
  rc = read(tcp->sock, buf + tcp->offset, bufSize - tcp->offset);
  if (rc > 0) {			/* Normal case */
    tcp->size += rc;
    if (tcp->size) {
      (*func)(ClientData, nextState);
      return(1);
    }
  } else if (rc == 0) {		/* EOF */
    if (compress) {
      fprintf(stderr, "WARNING! Got EOF on compressed stream\n");
     (*func)(ClientData, abortState);
     return(1);
    } else {
      fprintf(stderr, "Got EOF on stream\n");
      (*func)(ClientData, doneState);
      return(1);
    }
  } else if (rc == -1) {	/* Other error */
    if (WouldBlock(errno))
      return(0);		/* Caller should return, data is delayed */
    else if (errno == EINTR) {
      errno = 0;
      goto read_retry;		/* Probably a profiling timer */
    } else {
      fprintf(stderr,"Receive failed [TCPRead:RoverTCP.c]\n");
      (*func)(ClientData, abortState);
      return(1);
    }
  }
  (*func)(ClientData, nextState);
  return(1);
}

/* Gets a buffer in a non-blocking manner:
   Returns 1 when the state is changed.
   Returns 0 if more data is needed.
 */
PUBLIC int TCPGetBuffer(NBTcp *tcp, int *rc, StateChangeFunc func,
			void *ClientData, void *nextState, void *abortState)
{
  *rc = readbuffer(tcp->sock, tcp->line, 
		   (BUFSIZE > tcp->size) ? tcp->size : BUFSIZE);

  switch (*rc) {
  case -1:		/* EOF */
    *rc = 0;
    (*func)(ClientData, nextState);
    return(1);
    break;
  case 0:		/* No more data for now, keep going */
    return(0);
    break;
  case -2:		/* An error occurred (other than EINTR or EAGAIN) */
    fprintf(stderr, "Readbuffer error [TCPGetBuffer:RoverTCP.c]\n");
    *rc = 0;
    (*func)(ClientData, abortState);
    return(1);
    break;
  default:		/* Read some number of bytes */
    break;
  }
  return(0);
}

/* Write a buffer in a non-blocking manner:
   Returns 1 when the state is changed.
   Returns 0 if more data is needed.
 */
PUBLIC int TCPPutBuffer(NBTcp *tcp, StateChangeFunc func,
			void *ClientData, void *nextState, void *abortState)
{
  int rc = writebuffer(tcp->sock, tcp->ptr, &(tcp->offset), &(tcp->size));

  switch (rc) {
  case PUT_DONE:	/* Buffer completely written */
    (*func)(ClientData, nextState);
    return(1);
    break;
  case PUT_CONTINUE:	/* More data to send, keep going */
    return(0);
    break;
  case PUT_ERROR:	/* An error occurred (other than EINTR or EAGAIN) */
    fprintf(stderr, "Writebuffer error [TCPPutBuffer:RoverTCP.c]\n");
    (*func)(ClientData, abortState);
    return(1);
    break;
  case GOT_EOF:	/* EOF occurred */
    fprintf(stderr, "Writeline EOF [TCPPutBuffer:RoverTCP.c]\n");
    (*func)(ClientData, abortState);
    return(1);
    break;
  }
  (*func)(ClientData, abortState);
  return(1);
}

PUBLIC NBTcp *RoverTCPInit(int sock, int size)
{
  NBTcp *tcp = calloc(sizeof(NBTcp), 1);

  tcp->sock = sock;
  tcp->line = malloc(BUFSIZE);
  tcp->line[0] = '\0';
  tcp->offset = tcp->size = 0;
  return(tcp);
}

PUBLIC void RoverTCPFree(NBTcp *tcp)
{
  free(tcp->line);
  free(tcp);
}

/* Set a socket to close on exec */
PUBLIC int RoverSetCloseExec(int sockfd)
{
  return(fcntl(sockfd, F_SETFD, 1));
}

/* Make a socket non-blocking. If connected is 1, the procedure sets the
 *   O_NDELAY bit for DecStations running Ultrix.
 */
PUBLIC int RoverMakeSockNB(int sockfd, int connected)
{
  /* If non-blocking protocol then change socket status
   ** I use FCNTL so that I can ask the status before I set it.
   ** See W. Richard Stevens (Advan. Prog. in UNIX environment, p.364)
   ** Be CAREFULL with the old `O_NDELAY' - it will not work as read()
   ** returns 0 when blocking and NOT -1. FNDELAY is ONLY for BSD and
   ** does NOT work on SVR4 systems. O_NONBLOCK is POSIX.
   */

  int status;

#if defined(_WINDOWS) || defined(VMS)
  {
    int enable = 1;		/* Need the variable! */
    status = IOCTL(sockfd, FIONBIO, enable);
  }
#else
  {
#ifdef decstation
    /* Ultrix wants to have the O_NDELAY bit set on the socket.
       Why set it here and not before? Because the connect will
       return would block and that is an error. (Shouldn't there
       be a select or better yet an FSM for link state, similar
       to the one for HTTP state).
       
       From man 2 read:
     
       When attempting to read from an empty pipe (or FIFO):
     
       +    If no process has the pipe open for writing, read
            returns zero to indicate end-of-file.
	  
       +    If some process has the pipe open for writing
            and O_NDELAY or O_NONBLOCK is set, read returns a
	    -1, errno is to [EWOULDBLOCK]. If some process has
	    the pipe open for writing and O_NDELAY and O_NONBLOCK
	    are clear, read blocks until data is written or the
	    pipe is closed by all processes that opened the pipe
	    for writing.
     */
    int flag = !connected ? O_NONBLOCK : (O_NDELAY|O_NONBLOCK);	/* POSIX */
#else
    int flag = O_NONBLOCK;					/* POSIX */
#endif /* decstation */
    if ((status = fcntl(sockfd, F_GETFL, 0)) != -1) {
      status |= flag;
      status = fcntl(sockfd, F_SETFL, status);
    }
  }
#endif
  return(status);
}

PUBLIC int RoverTCPAccept(int socket, int execClose)
{
#ifdef linux
  struct sockaddr sa_client;
#else
  struct sockaddr_in sa_client;
#endif
  int sock, clen = sizeof(sa_client);

  /* Wait for socket connection (shouldn't have to wait at all since
   *   this is supposed to be called when the select loop calls us).
   */
  sock = accept(socket, &sa_client, &clen);
  if (sock < 0) { 	/* I don't think this should ever happen.  */
    perror("Error while attempting to accept Rover httpd request");
    return(-1);
  }
  if (RoverDebug)
    fprintf(stderr, "Socket accept %d [RoverTCPAccept:RoverTCP.c]\n", sock);
  RoverSetCloseExec(sock);	/* Set to close on exec */
  RoverMakeSockNB(sock, 1);	/* Don't want to block on I/O operations */
  return(sock);
}

PUBLIC void RoverGetRemoteInfo(int fd, char **remoteName, char **remoteIP)
{
  struct sockaddr addr;
  int len;
  struct in_addr *iaddr;
  struct hostent *hptr;

  len = sizeof(struct sockaddr);
  FREE(*remoteName);
  FREE(*remoteIP);

  if ((getpeername(fd, &addr, &len)) < 0) {
    *remoteName = (char *) malloc(sizeof(char)*13);
    strcpy(*remoteName, "UNKNOWN_HOST");
    return;
  }

  iaddr = &(((struct sockaddr_in *)&addr)->sin_addr);
#ifndef MINIMAL_DNS
  hptr = gethostbyaddr((char *)iaddr, sizeof(struct in_addr), AF_INET);
  if (hptr) {
    *remoteName = strdup(hptr->h_name);
    str_tolower(*remoteName);
  } else if (RoverDebug) {
    fprintf(stderr, "Can't find internet node name for peer!! [RoverGetRemoteInfo:RoverTCP.c]\n");
  }
#endif

#ifdef MAXIMUM_DNS
  /* Grrr. Check THAT name to make sure it's really the name of the addr. */
  /* Code from Harald Hanche-Olsen <hanche@imf.unit.no> */
  if (remote_host) {
    char **haddr;
    
    hptr = gethostbyname(remote_host);
    if (hptr) {
      for (haddr=hptr->h_addr_list;*haddr;haddr++) {
	if (((struct in_addr *)(*haddr))->s_addr == iaddr->s_addr)
	  break;
      }
    }
    if ((!hptr) || (!(*haddr))) FREE(remote_host);
  }
#endif
  *remoteIP = strdup(inet_ntoa(*iaddr));
  if (!*remoteName){
    *remoteName = (char *) malloc (sizeof(char)*15);
    strcpy(*remoteName,"UNKNOWN_HOST");
  }
  
  if (RoverDebug) 
    fprintf(stderr,
	    "Remote IP: %s, remote name: %s [RoverGetRemoteInfo:RoverTCP.c]\n",
	    *remoteIP, *remoteName);
}




/* Apparently this stuff wasn't there on the Suns upstairs. */
/* extern struct hostent *gethostbyname(char *name); */
#ifndef INADDR_NONE
#define INADDR_NONE     0xffffffff       /* should be in <netinet/in.h> */
#endif
#define RETRIES 16		/* Retry HTTP connects many times. */

/*
 * Opens a TCP connection to hostname host, on port number port. Returns
 *   the socket descriptor on success, or -1 on failure. May print error
 *   messages to stderr (the error log).
 */
PUBLIC int openTCP(char *host, int port)
{
  int setup = 0, sockfd = -1, flag, try = 0, delay = 1;
  struct sockaddr_in serv_addr;
  struct hostent *hp;
  unsigned long inaddr;
  struct hostent tcp_host_info; /* from gethostbyname() */
  
  while (setup++ < RETRIES) {
    bzero((char *) &serv_addr, sizeof(struct sockaddr_in));
    if ((inaddr = inet_addr(host)) != INADDR_NONE) {
      /* It's dotted-decimal */
      bcopy((char *) &inaddr, (char *) &serv_addr.sin_addr, sizeof(inaddr));
      tcp_host_info.h_name = NULL;
    } else {
      if ((hp = gethostbyname(host)) == NULL) {
	perror("openTCP");
	continue;
      }
      tcp_host_info = *hp;        /* found it by name, structure copy */
      bcopy(hp->h_addr, (char *) &serv_addr.sin_addr, hp->h_length);
    }
    
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_port = htons(port);
    
    if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
      perror("openTCP");
      continue;
    }
    
    flag = 1;
    if (setsockopt(sockfd, IPPROTO_TCP, TCP_NODELAY, (char *) &flag,
		   sizeof(int)) < 0) perror("openTCP");
    flag = 1;
    if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (char *) &flag,
		   sizeof(int)) < 0) perror("openTCP");
    break;
  }

  while (1) {		/* Retry many times. */
    fprintf(stderr, "Trying to open %s:%d (delay: %d)\n", host, port, delay);
    if (try >= RETRIES) break;
    if (try++) {
      sleep(delay);
      if (delay < 8)
	delay++;	/* Linear backoff */
      else	
	delay <<= 1;	/* Exponential backoff */
    }

    if (connect(sockfd,(struct sockaddr *) &serv_addr,sizeof(serv_addr)) < 0) {
      perror("openTCP");
      close(sockfd);
      continue;
    } else {
      fprintf(stderr, "Socket opened\n");
      return(sockfd);
    }
  }
  return(-1);
}

extern int RoverTCPValidSocket(int socket)
{
  int rc = fcntl(socket, F_GETFD);
  if (rc == -1) return(0);
  else return(1);
}
