 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: RoverZlib.c
 *
 * Description: Interface to zlib compression library
 *
 *
 ***********************************************************************
 *   Copyright 1996
 *   Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "RoverBase.h"
#include "RoverZlib.h"

/* Return the string equivalent to a zlib return code */
PRIVATE char *zlibErr2String(int code)
{    
  switch (code) {
  case Z_OK:
    return("Z_OK");
    break;
  case Z_STREAM_END:
    return("Z_STREAM_END");
    break;
  case Z_NEED_DICT:
    return("Z_NEED_DICT");
    break;
  case Z_ERRNO:
    return("Z_ERRNO");
    break;
  case Z_STREAM_ERROR:
    return("Z_STREAM_ERROR");
    break;
  case Z_DATA_ERROR:
    return("Z_DATA_ERROR");
    break;
  case Z_MEM_ERROR:
    return("Z_MEM_ERROR");
    break;
  case Z_BUF_ERROR:
    return("Z_BUF_ERROR");
    break;
  case Z_VERSION_ERROR:
    return("Z_VERSION_ERROR");
    break;
  default:
    return("UNKNOWN");
    break;
  }
}

/* Initialize a uncompressor */
PUBLIC z_stream *InitUncompress()
{
  z_stream *inf;
  int rc;

  if ((inf = (z_stream *)calloc(1, sizeof(z_stream))) == NULL)
    return(NULL);

  inf->zalloc = (alloc_func) 0;
  inf->zfree = (free_func) 0;
  inf->opaque = (voidpf) 0;
  if ((rc = inflateInit(inf)) != Z_OK) {
    fprintf(stderr, "inflateInit err %s (%d) [InflateInit:RoverZlib.c]\n", 
	    zlibErr2String(rc), rc);
    free(inf);
    return(NULL);
  }
  return(inf);
}

/* 
 * Uncompress a block and, if supplied, exec a function on each output segment
 * Returns 0 if more input data is needed
 * Returns 1 if the input decompression is done
 * Returns 2 if the input decompression is done and input bytes remain
 * Returns -1 if an error occured
 */
PUBLIC int UncompressCore(z_stream *inf, char *in, int *inLen, char *out,
			  int *outLen, int outMax, int *written,
			  writeFunction fcn, void *data)
{
  int rc = Z_OK, size = 0, rv;
  char *start;

  *written = 0;
  inf->next_in = (Bytef *) in;
  inf->avail_in = *inLen;
  inf->avail_out = outMax - *outLen;
  inf->next_out = (Bytef *) (out + *outLen);

  if (RoverDebug) {
    fprintf(stderr,
	    "In %d/Avail %d, Out %d/Avail %d [UncompressCore:RoverZlib.c]\n",
	    *inLen, inf->avail_in, *outLen, inf->avail_out);
    fprintf(stderr, "NI 0x%x, NO 0x%x, *outLen %d\n",
	    (unsigned int) inf->next_in, (unsigned int) inf->next_out,*outLen);
  }
  while (inf->avail_in && (rc == Z_OK)) {
    /* Don't consider no space to be an error. We may have the EOD marker to
     *  process. Also, zlib itself will return an error if the output buffer
     *  is full
     */
#if 0
    if (inf->avail_out <= 0) {
      if (!inf->avail_in) return(0);
      fprintf(stderr, "Inflation output buffer overflow [UncompressCore:RoverZlib.c]\n");
      free(inf);
      return(-1);
    }
#endif
    rc = inflate(inf, Z_PARTIAL_FLUSH);
    if (RoverDebug) fprintf(stderr,
			    "Inflate rc: %s/%d [UncompressCore:RoverZlib.c]\n",
			    zlibErr2String(rc), rc);
    size = (int) ((char *) inf->next_out - (out + *outLen));
    start = out + *outLen;
    *outLen += size;
    if (rc < 0) break;
    if (*outLen != ((int) ((char *) inf->next_out - out))) {
      fprintf(stderr, "Inflated wrong number of bytes [UncompressCore:RoverZlib.c]\n");
      return(-1);
    }
    if (*outLen) {
      *written += size;
      if (RoverDebug) fprintf(stderr, "Inflated %d bytes [UncompressCore:RoverZlib.c]\n", size);
      if (fcn != NULL) {
	rv = fcn(data, out, start, size);
	if (rv < 0) {
	  fprintf(stderr,"writeFunction error %d [UncompressCore:RoverZlib.c]\n", rv);
	  free(inf);
	  return(-1);
	} else if (rv) {
	  inf->avail_out += rv;
	  inf->next_out = (Bytef *) (out + (*outLen - rv));
	  *outLen -= rv;
	  if (((char *)inf->next_out) < out) {
	    fprintf(stderr, "Inflate error [UncompressCore:RoverZlib.c]\n");
	    return(-1);
	  }
/*	  memmove((void *) out, (void *)(out + *outLen), *outLen); */
	}
      }
    }
  }
  *inLen = inf->avail_in;
  if (RoverDebug) {
    fprintf(stderr,
	    "After: In %d/Av %d, Out %d/Av %d [UncompressCore:RoverZlib.c]\n",
	    *inLen, inf->avail_in, *outLen, inf->avail_out);
    fprintf(stderr, "After: NI 0x%x, NO 0x%x\n", (unsigned int) inf->next_in,
	    (unsigned int) inf->next_out);
  }
  if (inf->avail_in) {	/* Copy down any leftover input data */
    if (RoverDebug) 
      fprintf(stderr, "Saving %d bytes after uncompress data end [UncompressCore:RoverZlib.c]\n",*inLen);
    memmove((void *) in, (void *)inf->next_in, *inLen);
  }
  if (rc == Z_STREAM_END) {
/* >>>>> Call function at end if data remains (*outLen != 0) */
    if (RoverDebug) fprintf(stderr, "Inflate: Input size: %ld Output size: %ld [UncompressCore:RoverZlib.c]\n", inf->total_in, inf->total_out);
    rc = inflateEnd(inf);
    free(inf);
    if (inf->avail_in) return(2);
    else return(1);
  } else if (rc != Z_OK) {
    fprintf(stderr,"Inflate error %s (%d) [UncompressCore:RoverZlib.c]\n",
	    zlibErr2String(rc), rc);
    free(inf);
    return(-1);
  }
  return(0);	/* More input data needed */
}

PRIVATE int writeBuffer(void *data, char *outStart, char *out, int newSize)
{
  if (RoverDebug)
    fprintf(stderr, "Writing %d from 0x%x, `%s'\n", newSize,
	    (unsigned int) out, out);
  return(fwrite(out, 1, newSize, (FILE *) data));
}

/* 
 * Uncompress a block and write it to a stream
 * Returns 0 if more input data is needed
 * Returns 1 if the input decompression is done
 * Returns 2 if the input decompression is done and input bytes remain
 * Returns -1 if an error occured
 */
PUBLIC int Uncompress(z_stream *inf, FILE *sock, char *in, int *inLen,
		      char *out, int *outLen, int outMax, int *written)
{
  return(UncompressCore(inf, in, inLen, out, outLen, outMax, written,
			writeBuffer, (void *) sock));
}

/* 
 * Uncompress a block into a buffer
 * Returns 0 if more input data is needed
 * Returns 1 if the input decompression is done
 * Returns 2 if the input decompression is done and input bytes remain
 * Returns -1 if an error occured
 */
PUBLIC int UncompressBuffer(z_stream *inf, char *in, int *inLen, char *out,
			    int *outLen, int outMax, int *written)
{
  return(UncompressCore(inf, in, inLen, out, outLen, outMax, written,
			(writeFunction)NULL, (void *) NULL));
}

/* End the compressing of a stream of data */
PUBLIC int EndUncompress(z_stream *inf)
{
  inflateEnd(inf);
  if (inf->avail_in) {
    if (RoverDebug) 
      fprintf(stderr, "Remaining data (%d bytes) on compressed stream? [EndUncompress:RoverZlib.c]\n",
	      inf->avail_in);
    return(1);
  } else { 
    if (RoverDebug) fprintf(stderr, "Done uncompressing data [EndUncompress:RoverZlib.c]\n");
  }
  free(inf);
  return(0);
}

/* Initialize a compressor */
PUBLIC z_stream *InitCompress(int compressLevel)
{
  z_stream *c_stream;
  int rc;

  if ((c_stream = (z_stream *)calloc(1, sizeof(z_stream))) == NULL)
    return(NULL);

  c_stream->zalloc = (alloc_func) 0;
  c_stream->zfree = (free_func) 0;
  c_stream->opaque = (voidpf) 0;
  if ((rc = deflateInit(c_stream, compressLevel)) != Z_OK) {
    fprintf(stderr, "deflateInit error %d \n", rc);
    free(c_stream);
    return(NULL);
  }
  return(c_stream);
}

/* Compress a block and write it to a stream */
PUBLIC int Compress(z_stream *c_stream, FILE *sock, char *in, int inLen,
		    char *out, int outLen, int *written)
{
  int rc = Z_OK, size;
  
  *written = 0;
  c_stream->next_in = (Bytef *) in;
  c_stream->avail_in = inLen;
  c_stream->next_out = (Bytef *) out;
  c_stream->avail_out = outLen;
  while (c_stream->avail_in) {
    if ((rc = deflate(c_stream, Z_NO_FLUSH)) != Z_OK) {
      fprintf(stderr, "deflate error %d \n", rc);
      free(c_stream);
      return(rc);
    }
    if (c_stream->avail_out != outLen) {
      size = outLen - c_stream->avail_out;
      *written += size;
      fwrite(out, 1, size, sock);
      c_stream->next_out = (Bytef *) out;
      c_stream->avail_out = outLen;
    }
  }
  return(rc);
}

/* End the compressing of a stream of data */
PUBLIC int EndCompress(z_stream *c_stream, FILE *sock, char *out, int outLen,
		       int *written)
{
  int rc = Z_OK, size;

  *written = 0;
  while (rc != Z_STREAM_END) {
    rc = deflate(c_stream, Z_FINISH);
    if ((rc != Z_OK) && (rc != Z_STREAM_END)) {
      fprintf(stderr, "deflate error %d \n", rc);
      free(c_stream);
      return(rc);
    }
    if (c_stream->avail_out != outLen) {
      size = outLen - c_stream->avail_out;
      *written += size;
      fwrite(out, 1, size, sock);
      c_stream->next_out = (Bytef *) out;
      c_stream->avail_out = outLen;
    }
  }

  if (RoverDebug) 
    fprintf(stderr, "Deflate: Input size: %ld Output size: %ld\n",
	    c_stream->total_in, c_stream->total_out);
  if ((rc = deflateEnd(c_stream)) != Z_OK) {
    fprintf(stderr, "deflateEnd error %d \n", rc);
    free(c_stream);
    return(rc);
  }
  free(c_stream);
  return(rc);
}

