 /**********************************************************************
 *
 *
 *		    ROVER object execution engine
 *			  Anthony D. Joseph
 *		   Laboratory for Computer Science
 *		Massachusetts Institute of Technology
 *
 * Module: html.c
 *
 * Description: Brutally simple HTML parser for browser interface
 *		 layer for ROVER
 *
 * Last Modified:  28-June-95 (adj)
 *
 * Global Functions:
 *    main
 *
 * Global Variables: none
 *
 ***********************************************************************
 *   Copyright 1995
 *   Alan deLespinasse and Anthony D. Joseph
 *   Massachusetts Institute of Technology
 *
 *   Permission to use, copy, modify, and distribute this program
 *   for any purpose and without fee is hereby granted, provided
 *   that this copyright and permission notice appear on all copies
 *   and supporting documentation, the name of M.I.T. not be used
 *   in advertising or publicity pertaining to distribution of the
 *   program without specific prior permission, and notice be given
 *   in supporting documentation that copying and distribution is
 *   by permission of M.I.T.  M.I.T. makes no representations about
 *   the suitability of this software for any purpose.  It is pro-
 *   vided "as is" without express or implied warranty.		
 ***********************************************************************/

/* Can be compiled with -DSTANDALONE to produce test program. Just takes
   HTML on stdin and says what's text and what's tags. */

#include <stdlib.h>
#include "html.h"

/* Create a new parser. */
struct HTMLparser* HTMLnewparse(void (*text)(struct HTMLparser*),
				void (*tag)(struct HTMLparser*), void *d)
{
  struct HTMLparser *n = (struct HTMLparser*)
    malloc(sizeof(struct HTMLparser)+BUFSIZE);
  n->textcb = text;
  n->tagcb = tag;
  n->length = 0;
  n->incomment = n->intag = n->inquotes = n->backslashed = 0;
  n->data = d;
  return(n);
}

/* Parse a chunk of HTML. */
void HTMLparse(struct HTMLparser* p, char* s, size_t size)
{
  size_t n = 0;
  char c;

#ifdef HTMLDEBUG
  fprintf(stderr, "Enter HTMLparse (intag? %s, incomment? %s)\n",
	  p->intag ? "YES" : "NO", p->incomment ? "YES" : "NO");
  {
    int i;
    fprintf(stderr, "Got `");
    for (i = 0; i < size; i++)
      fprintf(stderr, "%c", s[i]);
    fprintf(stderr, "'\n");
  }
#endif /* HTMLDEBUG */

  for (;;) {
    if (!p->incomment && !p->intag) {	/* not inside <>'s */
      for (;;) {			/* parse plain text */
	if (n >= size) return;		/* done parsing chunk */
	c = s[n];			/* next char */
	if (!p->backslashed) {		/* backslash means literal < */
	  if (c == '<') {		/* starting a tag */
	    p->buf[p->length]=0; 	/* terminate string */
	    if (p->textcb) 		/* call callback */
	      (*(p->textcb))(p);
	    p->buf[0] = '<';
	    p->length = 1;		/* start the tag off */
	    p->intag = 1;
	    n++;
	    break;			/* and jump to tag-parsing loop */
	  }
	  if (c == '\\') p->backslashed = 1;
	} else p->backslashed = 0;
	p->buf[p->length] = c;		/* add char to buf */
	p->length++;
	n++;
	if ((p->length+1) >= BUFSIZE) { /* don't overflow buffer */
	  /* Should reallocate buffer here !!! */
	  p->buf[p->length] = 0;
	  if (p->textcb) (*(p->textcb))(p);
	  p->length = 0;
	}
      }
    }
    
    for (;;) {				/* parse tag */
      if (n >= size) return;
      c = s[n];
      p->buf[p->length] = c;
      p->length++;
      n++;
      if (p->length == 4) {
#ifdef HTMLDEBUG
	fprintf(stderr, "Comment? ");
#endif /* HTMLDEBUG */
	if ((p->buf[1] == '!') && (p->buf[2] == '-') && (p->buf[3] == '-')) {
#ifdef HTMLDEBUG
	  fprintf(stderr, "YES\n");
#endif /* HTMLDEBUG */
	  p->intag = 0;
	  p->incomment = 1;
	}
#ifdef HTMLDEBUG
	else fprintf(stderr, "NO\n");
#endif /* HTMLDEBUG */
      }
      if (!p->backslashed) {		/* backslash means ignore next char */
	if (!p->inquotes) {	/* double quotes contain literal strings */
	  if (c == '>') {		/* end of tag? */
#ifdef HTMLDEBUG
	    fprintf(stderr, "End");
#endif /* HTMLDEBUG */
	    if (p->intag) { 
#ifdef HTMLDEBUG
	      fprintf(stderr, " tag\n");
#endif /* HTMLDEBUG */
	      break;	/* end of tag */
	    } else if ((p->length > 7) &&
		       (p->buf[p->length-3] == '-') &&
		       (p->buf[p->length-2] == '-')) {	/* end of comment */
#ifdef HTMLDEBUG
	      fprintf(stderr, " comment\n");
#endif /* HTMLDEBUG */
	      break;
	    }
#ifdef HTMLDEBUG
	    else fprintf(stderr, " comment FAIL\n");
#endif /* HTMLDEBUG */
	  }
	}
	if (c == '\"') p->inquotes = !p->inquotes;
	else if (c == '\\') p->backslashed = 1;
      } else (p->backslashed = 0);
    }
    if (p->intag) {
      p->buf[p->length] = 0;
      if (p->tagcb) (*(p->tagcb))(p); 	/* call callback */
    }
    p->length = 0;
    p->incomment = p->intag = 0;	/* start over with plain text */
  }
}

/* Finish parse job and deallocate parser. */
void HTMLendparse(struct HTMLparser* p)
{
  if (p->length) {
    p->buf[p->length] = 0;
    if (p->intag) {
      if (p->tagcb) (*(p->tagcb))(p);
    } else {
      if (p->textcb) (*(p->textcb))(p);
    }
  }
  free(p);
}

#ifdef STANDALONE

void textcallb(struct HTMLparser* p)
{
  fprintf(stderr, "Text: %s\n", p->buf);
}

void tagcallb(struct HTMLparser* p)
{
  fprintf(stderr, "Tag: %s\n", p->buf);
}

main() {
  int n;
  char buf[50];
  struct HTMLparser *p = HTMLnewparse(textcallb, tagcallb, NULL);
  while (n = fread(buf, 1, 50, stdin)) HTMLparse(p, buf, n);
  HTMLendparse(p);
  return 0;
}

#endif /* STANDALONE */
